"""
    SoftLayer.file
    ~~~~~~~~~~~~~~~
    File Storage Manager

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer.managers.storage import StorageManager
from SoftLayer.managers import storage_utils
from SoftLayer import utils

# pylint: disable=too-many-public-methods


class FileStorageManager(StorageManager):
    """Manages file Storage volumes."""

    def list_file_volume_limit(self):
        """Returns a list of block volume count limit.

        :return: Returns a list of block volume count limit.
        """
        return self.get_volume_count_limits()

    def list_file_volumes(self, datacenter=None, username=None, storage_type=None, order=None, **kwargs):
        """Returns a list of file volumes.

        :param datacenter: Datacenter short name (e.g.: dal09)
        :param username: Name of volume.
        :param storage_type: Type of volume: Endurance or Performance
        :param order: Volume order id.
        :param kwargs:
        :return: Returns a list of file volumes.
        """
        if 'mask' not in kwargs:
            items = [
                'id',
                'username',
                'capacityGb',
                'bytesUsed',
                'serviceResource.datacenter[name]',
                'serviceResourceBackendIpAddress',
                'activeTransactionCount',
                'fileNetworkMountAddress',
                'replicationPartnerCount'
            ]
            kwargs['mask'] = ','.join(items)

        _filter = utils.NestedDict(kwargs.get('filter') or {})
        _filter['nasNetworkStorage']['id'] = utils.query_filter_orderby()
        _filter['nasNetworkStorage']['serviceResource']['type']['type'] = utils.query_filter('!~ NAS')

        _filter['nasNetworkStorage']['storageType']['keyName'] = (
            utils.query_filter('*FILE_STORAGE*'))
        if storage_type:
            _filter['nasNetworkStorage']['storageType']['keyName'] = (
                utils.query_filter('%s_FILE_STORAGE*' % storage_type.upper()))

        if datacenter:
            _filter['nasNetworkStorage']['serviceResource']['datacenter'][
                'name'] = utils.query_filter(datacenter)

        if username:
            _filter['nasNetworkStorage']['username'] = utils.query_filter(username)

        if order:
            _filter['nasNetworkStorage']['billingItem']['orderItem'][
                'order']['id'] = utils.query_filter(order)

        kwargs['filter'] = _filter.to_dict()
        return self.client.call('Account', 'getNasNetworkStorage', iter=True, **kwargs)

    def get_file_volume_details(self, volume_id, **kwargs):
        """Returns details about the specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns details about the specified volume.
        """

        if 'mask' not in kwargs:
            items = [
                'id',
                'username',
                'password',
                'capacityGb',
                'bytesUsed',
                'snapshotCapacityGb',
                'parentVolume.snapshotSizeBytes',
                'storageType.keyName',
                'serviceResource.datacenter[name]',
                'serviceResourceBackendIpAddress',
                'fileNetworkMountAddress',
                'storageTierLevel',
                'provisionedIops',
                'lunId',
                'originalVolumeName',
                'originalSnapshotName',
                'originalVolumeSize',
                'activeTransactionCount',
                'activeTransactions.transactionStatus[friendlyName]',
                'replicationPartnerCount',
                'replicationStatus',
                'replicationPartners[id,username,'
                'serviceResourceBackendIpAddress,'
                'serviceResource[datacenter[name]],'
                'replicationSchedule[type[keyname]]]',
                'notes',
            ]
            kwargs['mask'] = ','.join(items)
        return self.get_volume_details(volume_id, **kwargs)

    def get_file_volume_access_list(self, volume_id, **kwargs):
        """Returns a list of authorized hosts for a specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns a list of authorized hosts for a specified volume.
        """
        return self.get_volume_access_list(volume_id, **kwargs)

    def get_file_volume_snapshot_list(self, volume_id, **kwargs):
        """Returns a list of snapshots for the specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns a list of snapshots for the specified volume.
        """
        return self.get_volume_snapshot_list(volume_id, **kwargs)

    def order_file_volume(self, storage_type, location, size,
                          iops=None, tier_level=None, snapshot_size=None,
                          service_offering='storage_as_a_service',
                          hourly_billing_flag=False):
        """Places an order for a file volume.

        :param storage_type: 'performance' or 'endurance'
        :param location: Name of the datacenter in which to order the volume
        :param size: Size of the desired volume, in GB
        :param iops: Number of IOPs for a "Performance" order
        :param tier_level: Tier level to use for an "Endurance" order
        :param snapshot_size: The size of optional snapshot space,
            if snapshot space should also be ordered (None if not ordered)
        :param service_offering: Requested offering package to use in the order
            ('storage_as_a_service', 'enterprise', or 'performance')
        :param hourly_billing_flag: Billing type, monthly (False)
            or hourly (True), default to monthly.
        """
        order = storage_utils.prepare_volume_order_object(
            self, storage_type, location, size, iops, tier_level,
            snapshot_size, service_offering, 'file', hourly_billing_flag
        )

        return self.client.call('Product_Order', 'placeOrder', order)

    def cancel_file_volume(self, volume_id, reason='No longer needed', immediate=False):
        """Cancels the given file storage volume.

        :param integer volume_id: The volume ID
        :param string reason: The reason for cancellation
        :param boolean immediate: Cancel immediately or on anniversary date
        """
        return self.cancel_volume(volume_id, reason, immediate)

    def _get_ids_from_username(self, username):
        object_mask = "mask[id]"
        results = self.list_file_volumes(username=username, mask=object_mask)
        if results:
            return [result['id'] for result in results]
        return []



"""List firewalls."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@environment.pass_env
def cli(env):
    """List firewalls."""

    mgr = SoftLayer.FirewallManager(env.client)
    table = formatting.Table(['firewall id',
                              'type',
                              'features',
                              'server/vlan id'], title='Single Server Firewalls')
    fwvlans = mgr.get_firewalls()
    dedicated_firewalls = [firewall for firewall in fwvlans
                           if firewall['dedicatedFirewallFlag']]

    for vlan in dedicated_firewalls:
        features = []
        if vlan['highAvailabilityFirewallFlag']:
            features.append('HA')

        if features:
            feature_list = formatting.listing(features, separator=',')
        else:
            feature_list = formatting.blank()

        table.add_row([
            'vlan:%s' % vlan['networkVlanFirewall']['id'],
            'VLAN - dedicated',
            feature_list,
            vlan['id']
        ])

    shared_vlan = [firewall for firewall in fwvlans
                   if not firewall['dedicatedFirewallFlag']]
    for vlan in shared_vlan:
        vs_firewalls = [guest
                        for guest in vlan['firewallGuestNetworkComponents']
                        if has_firewall_component(guest)]

        for firewall in vs_firewalls:
            table.add_row([
                'vs:%s' % firewall['id'],
                'Virtual Server - standard',
                '-',
                firewall['guestNetworkComponent']['guest']['id']
            ])

        server_firewalls = [server
                            for server in vlan['firewallNetworkComponents']
                            if has_firewall_component(server)]

        for firewall in server_firewalls:
            table.add_row([
                'server:%s' % firewall['id'],
                'Server - standard',
                '-',
                utils.lookup(firewall,
                             'networkComponent',
                             'downlinkComponent',
                             'hardwareId')
            ])

    table_gatewalls = formatting.Table(['Id',
                                        'firewall',
                                        'type',
                                        'Hostname',
                                        'Location',
                                        'Public Ip',
                                        'Private Ip',
                                        'Associated vlan',
                                        'status'], title='Multi Vlan Firewall')
    fw_gatewwalls = mgr.get_firewalls_gatewalls()

    for gatewalls in fw_gatewwalls:
        table_gatewalls.add_row([gatewalls['networkFirewall']['id'], gatewalls.get('name'),
                                 gatewalls['networkFirewall']['firewallType'],
                                 gatewalls['members'][0]['hardware']['hostname'],
                                 gatewalls['networkFirewall']['datacenter']['name'],
                                 gatewalls['publicIpAddress']['ipAddress'],
                                 gatewalls['privateIpAddress']['ipAddress'],
                                 len(gatewalls['insideVlans']), gatewalls['status']['keyName']])
    env.fout(table)
    env.fout(table_gatewalls)


def has_firewall_component(server):
    """Helper to determine whether or not a server has a firewall.

    :param dict server: A dictionary representing a server
    :returns: True if the Server has a firewall.
    """
    if server['status'] != 'no_edit':
        return True

    return False



"""
    SoftLayer.transports.timing
    ~~~~~~~~~~~~~~~~~~~~
    Timing transport, used when you want to know how long an API call took.

    :license: MIT, see LICENSE for more details.
"""
import time


class TimingTransport(object):
    """Transport that records API call timings."""

    def __init__(self, transport):
        self.transport = transport
        self.last_calls = []

    def __call__(self, call):
        """See Client.call for documentation."""
        start_time = time.time()

        result = self.transport(call)

        end_time = time.time()
        self.last_calls.append((call, start_time, end_time - start_time))
        return result

    def get_last_calls(self):
        """Retrieves the last_calls property.

        This property will contain a list of tuples in the form
        (Request, initiated_utc_timestamp, execution_time)
        """
        last_calls = self.last_calls
        self.last_calls = []
        return last_calls

    @staticmethod
    def print_reproduceable(call):
        """Not Implemented"""
        return call.service



"""List Object Storage endpoints."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """List object storage endpoints."""

    mgr = SoftLayer.ObjectStorageManager(env.client)
    endpoints = mgr.list_endpoints(identifier)

    final_end_points = []

    table = formatting.Table(['Location/Region', 'Url', 'EndPoint Type', 'Public/Private', 'Legacy'])
    table.align['Location/Region'] = 'l'
    table.align['Url'] = 'l'
    for endpoint in endpoints:
        data = {
            'Location/Region': location_region(endpoint),
            'Url': endpoint['url'],
            'EndPoint Type': end_point_return(endpoint['region']),
            'Public/Private': public_private(endpoint['type']),
            'Legacy': endpoint['legacy']
        }
        final_end_points.append(data)

    final_end_points = sort_endpoint(final_end_points)
    table = add_array_to_table(table, final_end_points)

    env.fout(table)


def add_array_to_table(table, array_datas):
    """Add an array to a table"""
    for array in array_datas:
        table.add_row([array['Location/Region'],
                       array['Url'],
                       array['EndPoint Type'],
                       array['Public/Private'],
                       array['Legacy']])
    return table


def end_point_return(endpoint):
    """Returns end point type"""
    if endpoint == 'singleSite':
        return 'Single Site'
    if endpoint == 'regional':
        return 'Region'
    return 'Cross Region'


def public_private(data):
    """Returns public or private in capital letter"""
    if data == 'public':
        return 'Public'
    return 'Private'


def location_region(endpoint):
    """Returns location if it exists otherwise region"""
    if utils.lookup(endpoint, 'location'):
        return endpoint['location']
    return endpoint['region']


def sort_endpoint(endpoints):
    """Sort the all endpoints for public or private"""
    first_data = 0
    endpoint_type = ''
    if len(endpoints) > 0:
        endpoint_type = endpoints[first_data]['EndPoint Type']
    public = []
    private = []
    array_final = []
    for endpoint in endpoints:
        if endpoint['EndPoint Type'] != endpoint_type:
            endpoint_type = endpoint['EndPoint Type']
            array_final = array_final + public + private
            public.clear()
            private.clear()
        if endpoint['Public/Private'] == 'Public':
            public.append(endpoint)
        else:
            private.append(endpoint)

    array_final = array_final + public + private
    return array_final



"""Modifies a password for a volume's access"""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('access_id')
@click.option('--password', '-p', multiple=False,
              help='Password you want to set, this command will fail if the password is not strong.')
@environment.pass_env
def cli(env, access_id, password):
    """Changes a password for a volume's access.

    access id is the allowed_host_id from slcli block access-list
    """

    block_manager = SoftLayer.BlockStorageManager(env.client)

    result = block_manager.set_credential_password(access_id=access_id, password=password)

    if result:
        click.echo('Password updated for %s' % access_id)
    else:
        click.echo('FAILED updating password for %s' % access_id)



"""Add and upload SSL certificate details."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--crt', required=True, type=click.Path(exists=True), help="Certificate file")
@click.option('--key', type=click.Path(exists=True), required=True, help="Private Key file")
@click.option('--csr', type=click.Path(exists=True), help="Certificate Signing Request file")
@click.option('--icc', type=click.Path(exists=True), help="Intermediate Certificate file")
@click.option('--notes', help="Additional notes")
@environment.pass_env
def cli(env, crt, csr, icc, key, notes):
    """Add and upload SSL certificate details."""

    template = {
        'intermediateCertificate': '',
        'certificateSigningRequest': '',
        'notes': notes,
    }
    with open(crt, encoding="utf-8") as file_crt:
        template['certificate'] = file_crt.read()
    with open(key, encoding="utf-8") as file_key:
        template['privateKey'] = file_key.read()
    if csr:
        with open(csr, encoding="utf-8") as file_csr:
            body = file_csr.read()
            template['certificateSigningRequest'] = body

    if icc:
        with open(icc, encoding="utf-8") as file_icc:
            body = file_icc.read()
            template['intermediateCertificate'] = body

    manager = SoftLayer.SSLManager(env.client)
    manager.add_certificate(template)



"""List packages."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import ordering

COLUMNS = ['id',
           'name',
           'keyName',
           'type']


@click.command(cls=SLCommand)
@click.option('--keyword', help="A word (or string) used to filter package names.")
@click.option('--package_type', help="The keyname for the type of package. BARE_METAL_CPU for example")
@environment.pass_env
def cli(env, keyword, package_type):
    """List packages that can be ordered via the placeOrder API.


    ::

        # List out all packages for ordering
        slcli order package-list

        # List out all packages with "server" in the name
        slcli order package-list --keyword server

        # Select only specifict package types
        slcli order package-list --package_type BARE_METAL_CPU
    """
    manager = ordering.OrderingManager(env.client)
    table = formatting.Table(COLUMNS)

    _filter = {'type': {'keyName': {'operation': '!= BLUEMIX_SERVICE'}}}
    if keyword:
        _filter['name'] = {'operation': '*= %s' % keyword}
    if package_type:
        _filter['type'] = {'keyName': {'operation': package_type}}

    packages = manager.list_packages(filter=_filter)

    for package in packages:
        table.add_row([
            package['id'],
            package['name'],
            package['keyName'],
            package['type']['keyName']
        ])
    env.fout(table)



getObject = {
    'accountId': 1234,
    'backendRouterId': 1411193,
    'backendRouter': {
        'fullyQualifiedDomainName': 'bcr02a.dal13.softlayer.com',
        'hostname': 'bcr02a.dal13',
        'id': 1411193,
        'datacenter': {
            'id': 1854895,
            'longName': 'Dallas 13',
            'name': 'dal13',

        }
    },
    'createDate': '2018-09-24T16:33:09-06:00',
    'id': 3103,
    'modifyDate': '',
    'name': 'test-capacity',
    'instances': [
        {
            'createDate': '2018-09-24T16:33:09-06:00',
            'guestId': 62159257,
            'id': 3501,
            'billingItem': {
                'id': 348319479,
                'recurringFee': '3.04',
                'category': {'name': 'Reserved Capacity'},
                'item': {
                    'keyName': 'B1_1X2_1_YEAR_TERM'
                }
            },
            'guest': {
                'domain': 'cgallo.com',
                'hostname': 'test-reserved-instance',
                'id': 62159257,
                'modifyDate': '2018-09-27T16:49:26-06:00',
                'primaryBackendIpAddress': '10.73.150.179',
                'primaryIpAddress': '169.62.147.165'
            }
        },
        {
            'createDate': '2018-09-24T16:33:10-06:00',
            'guestId': 62159275,
            'id': 3519,
            'billingItem': {
                'id': 348319443,
                'recurringFee': '3.04',
                'category': {
                    'name': 'Reserved Capacity'
                },
                'item': {
                    'keyName': 'B1_1X2_1_YEAR_TERM'
                }
            }
        }
    ]
}


getObject_pending = {
    'accountId': 1234,
    'backendRouterId': 1411193,
    'backendRouter': {
        'fullyQualifiedDomainName': 'bcr02a.dal13.softlayer.com',
        'hostname': 'bcr02a.dal13',
        'id': 1411193,
        'datacenter': {
            'id': 1854895,
            'longName': 'Dallas 13',
            'name': 'dal13',

        }
    },
    'createDate': '2018-09-24T16:33:09-06:00',
    'id': 3103,
    'modifyDate': '',
    'name': 'test-capacity',
    'instances': [
        {
            'createDate': '2018-09-24T16:33:09-06:00',
            'guestId': 62159257,
            'id': 3501,
        }
    ]
}



"""
    SoftLayer.CLI.command
    ~~~~~~~~~~~~~~~~~~~~~
    Command interface for the SoftLayer CLI. Basically the Click commands, with fancy help text

    :license: MIT, see LICENSE for more details.
"""
import inspect
import types

import click

from rich import box
from rich.highlighter import RegexHighlighter
from rich.table import Table
from rich.text import Text

from SoftLayer.CLI import environment


class OptionHighlighter(RegexHighlighter):
    """Provides highlighter regex for the Command help.

    Defined in SoftLayer\\utils.py console_color_themes()
    """
    highlights = [
        r"(?P<switch>^\-\w)",  # single options like -v
        r"(?P<option>\-\-[\w\-]+)",  # long options like --verbose
        r"(?P<default_option>\[[^\]]+\])",  # anything between [], usually default options
        r"(?P<option_choices>Choices: )",
        r"(?P<example_block>Example::)",
        r"(?P<url>(file|https|http|ws|wss)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#~]*)"
        r"(?P<args_keyword>^[A-Z]+$)",
        r"(?P<deprecated>\(Deprecated\) .*$)"
    ]


class CommandLoader(click.MultiCommand):
    """Loads module for click."""

    def __init__(self, *path, **attrs):
        click.MultiCommand.__init__(self, **attrs)
        self.path = path
        self.highlighter = OptionHighlighter()
        self.env = None
        self.console = None

    def ensure_env(self, ctx):
        """ensures self.env is set"""
        if self.env is None:
            self.env = ctx.ensure_object(environment.Environment)
            self.env.load()
        if self.console is None:
            self.console = self.env.console

    def list_commands(self, ctx):
        """List all sub-commands."""
        self.ensure_env(ctx)
        return sorted(self.env.list_commands(*self.path))

    def get_command(self, ctx, cmd_name):
        """Get command for click."""
        self.ensure_env(ctx)
        # Do alias lookup (only available for root commands)
        if len(self.path) == 0:
            cmd_name = self.env.resolve_alias(cmd_name)

        new_path = list(self.path)
        new_path.append(cmd_name)
        module = self.env.get_command(*new_path)
        if isinstance(module, types.ModuleType):
            return CommandLoader(*new_path, help=module.__doc__ or '')
        else:
            return module

    # # pylint: disable=unused-argument
    def format_usage(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
        """Formats and colorizes the usage information."""
        self.ensure_env(ctx)
        pieces = self.collect_usage_pieces(ctx)
        for index, piece in enumerate(pieces):
            if piece == "[OPTIONS]":
                pieces[index] = "[options][OPTIONS][/]"
            elif piece == "COMMAND [ARGS]...":
                pieces[index] = "[command]COMMAND[/] [args][ARGS][/] ..."

        self.console.print(f"Usage: [path]{ctx.command_path}[/] {' '.join(pieces)}")

    # pylint: disable=unused-argument
    def format_help_text(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
        """Writes the help text"""
        text = self.help if self.help is not None else ""

        if self.deprecated:
            text = f"(Deprecated) {text}"

        if text:
            text = inspect.cleandoc(text).partition("\f")[0]
        self.console.print(f"\n\t{text}\n", highlight=True)

    def format_epilog(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
        """Writes the epilog if it exists, then prints out any sub-commands if they exist."""
        if self.epilog:
            epilog = inspect.cleandoc(self.epilog)
            epilog = epilog.replace("\n", " ")
            self.console.print(epilog)
        self.format_commands(ctx, formatter)

    # pylint: disable=unused-argument
    def format_options(self, ctx, formatter):
        """Prints out the options in a table format"""

        options_table = Table(highlight=True, box=box.SQUARE, show_header=False)

        for param in self.get_params(ctx):
            if len(param.opts) == 2:
                opt1 = self.highlighter(param.opts[1])
                opt2 = self.highlighter(param.opts[0])
            else:
                opt2 = self.highlighter(param.opts[0])
                opt1 = Text("")

            # Ensures the short option is always in opt1.
            if len(opt2) == 2:
                opt1, opt2 = opt2, opt1

            if param.metavar:
                opt2 += Text(f" {param.metavar}", style="bold yellow")

            options = Text(" ".join(reversed(param.opts)))
            help_record = param.get_help_record(ctx)
            help_message = ""
            if help_record:
                help_message = param.get_help_record(ctx)[-1]

            if param.metavar:
                options += f" {param.metavar}"
            options_table.add_row(opt1, opt2, self.highlighter(help_message))

        self.console.print(options_table)

    # pylint: disable=unused-argument
    def format_commands(self, ctx, formatter):
        """Formats the command list for click"""
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None or cmd.hidden:
                continue
            commands.append((subcommand, cmd))

        command_table = Table(highlight=True, box=None, show_header=False)
        if len(commands):
            for subcommand, cmd in commands:
                help_text = cmd.get_short_help_str(120)
                command_style = Text(f" {subcommand}", style="sub_command")
                command_table.add_row(command_style, help_text)
        self.console.print("\n[name_sub_command]Commands:[/]")
        self.console.print(command_table)


class SLCommand(click.Command):
    """Overloads click.Command to control how the help message is formatted."""

    def __init__(self, **attrs):
        click.Command.__init__(self, **attrs)
        self.highlighter = OptionHighlighter()
        self.env = None
        self.console = None

    def ensure_env(self, ctx):
        """ensures self.env is set"""
        if self.env is None:
            self.env = ctx.ensure_object(environment.Environment)
            self.env.load()
        if self.console is None:
            self.console = self.env.console

    def format_usage(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
        """Formats and colorizes the usage information."""
        self.ensure_env(ctx)
        pieces = self.collect_usage_pieces(ctx)
        for index, piece in enumerate(pieces):
            if piece == "[OPTIONS]":
                pieces[index] = "[options][OPTIONS][/]"
            elif piece == "COMMAND [ARGS]...":
                pieces[index] = "[command]COMMAND[/] [args][ARGS][/] ..."
            else:
                pieces[index] = f"[args_keyword]{piece}[/]"

        self.console.print(f"Usage: [path]{ctx.command_path}[/] {' '.join(pieces)}")

    def format_help_text(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
        """Writes the help text"""
        text = self.help if self.help is not None else ""

        if self.deprecated:
            text = f"(Deprecated) {text}"

        if text:
            text = f"\n\t{inspect.cleandoc(text)}\n"

        # Can't use F-string here because it messes with highlights
        self.console.print(self.highlighter(text))

    def format_epilog(self, ctx: click.Context, formatter: click.formatting.HelpFormatter) -> None:
        """Writes the epilog if it exists, then prints out any sub-commands if they exist."""
        if self.epilog:
            epilog = inspect.cleandoc(self.epilog)
            epilog = epilog.replace("\n", " ")
            self.console.print(epilog)

    def format_options(self, ctx, formatter):
        """Prints out the options in a table format"""

        options_table = Table(highlight=True, box=box.SQUARE, show_header=False)

        for param in self.get_params(ctx):
            # useful for showing whats in a param
            # print(param.to_info_dict())

            # Set Arguments to all uppercase
            if param.param_type_name == 'argument':
                param.opts[0] = param.opts[0].upper()

            # This option has a short (-v) and long (--verbose) options
            if len(param.opts) == 2:
                opt1 = self.highlighter(param.opts[1])
                opt2 = self.highlighter(param.opts[0])
            else:
                opt2 = self.highlighter(param.opts[0])
                # Needs to be the Text() type because rich.Text doesn't mesh with string
                opt1 = Text("")

            # Ensures the short option is always in opt1.
            if len(opt2) == 2:
                opt1, opt2 = opt2, opt1

            if param.metavar:
                opt2 += Text(f" {param.metavar}", style="bold yellow")

            # secondary_opts are usually for flags --enable/--disable
            if len(param.secondary_opts) == 1:
                opt2 += Text("|") + self.highlighter(param.secondary_opts[0])

            help_record = param.get_help_record(ctx)
            help_message = ""
            if help_record:
                help_message = Text(param.get_help_record(ctx)[-1])

            # Add Click choices to help message
            if isinstance(param.type, click.Choice):
                choices = ", ".join(param.type.choices)
                help_message += f" Choices: {choices}"

            options_table.add_row(opt1, opt2, self.highlighter(help_message))

        self.console.print(options_table)



verifyOrder = {
    'orderId': 1234,
    'orderDate': '2013-08-01 15:23:45',
    'prices': [{
        'id': 1,
        'laborFee': '2',
        'oneTimeFee': '2',
        'oneTimeFeeTax': '.1',
        'quantity': 1,
        'recurringFee': '2',
        'recurringFeeTax': '.1',
        'hourlyRecurringFee': '2',
        'setupFee': '1',
        'item': {'id': 1, 'description': 'this is a thing'},
    }]}
placeOrder = {
    'orderId': 1234,
    'orderDate': '2013-08-01 15:23:45',
    'orderDetails': {
        'prices': [{
            'id': 1,
            'laborFee': '2',
            'oneTimeFee': '2',
            'oneTimeFeeTax': '.1',
            'quantity': 1,
            'recurringFee': '2',
            'recurringFeeTax': '.1',
            'hourlyRecurringFee': '2',
            'setupFee': '1',
            'item': {'id': 1, 'description': 'this is a thing'},
        }],
        'virtualGuests': [{
            'id': 1234567,
            'globalIdentifier': '1a2b3c-1701',
            'fullyQualifiedDomainName': 'test.guest.com'
        }]
    }
}

# Reserved Capacity Stuff

rsc_verifyOrder = {
    'orderContainers': [
        {
            'locationObject': {
                'id': 1854895,
                'longName': 'Dallas 13',
                'name': 'dal13'
            },
            'name': 'test-capacity',
            'postTaxRecurring': '0.32',
            'prices': [
                {
                    'item': {
                        'id': 1,
                        'description': 'B1.1x2 (1 Year ''Term)',
                        'keyName': 'B1_1X2_1_YEAR_TERM',
                    }
                }
            ]
        }
    ],
    'postTaxRecurring': '0.32',
}

hardware_verifyOrder = {
    "currencyShortName": "USD",
    "hardware": [
        {
            "accountId": 1111,
            "domain": "testedit.com",
            "hostname": "test",
            "globalIdentifier": "81434794-af69-44d5-bb97-12312asdasdasd"
        }
    ],
    "location": "1441195",
    "locationObject": {
        "id": 1441195,
        "longName": "Dallas 10",
        "name": "dal10"
    },
    "packageId": 911,
    "postTaxRecurring": "0",
    "postTaxRecurringHourly": "0",
    "postTaxRecurringMonthly": "0",
    "preTaxRecurring": "0",
    "preTaxRecurringHourly": "0",
    "preTaxRecurringMonthly": "0",
    "prices": [
        {
            "hourlyRecurringFee": "0",
            "id": 209391,
            "recurringFee": "0",
            "categories": [
                {
                    "categoryCode": "ram",
                    "id": 3,
                    "name": "RAM"
                }
            ],
            "item": {
                "capacity": "32",
                "description": "32 GB RAM",
                "id": 11291,
                "units": "GB"
            }
        }
    ],
    "proratedInitialCharge": "0",
    "proratedOrderTotal": "0",
    "quantity": 1,
    "sendQuoteEmailFlag": None,
    "totalRecurringTax": "0",
    "useHourlyPricing": False
}

hardware_placeOrder = {
    "orderDate": "2021-05-07T07:41:41-06:00",
    "orderDetails": {
        "billingInformation": {
            "billingAddressLine1": "4849 Alpha Rd",
            "billingCity": "Dallas",
            "billingCountryCode": "US",
            "billingEmail": "test.ibm.com",
            "billingNameCompany": "SoftLayer Internal - Development Community",
            "billingNameFirst": "Test",
            "billingNameLast": "Test",
            "billingPhoneVoice": "1111111",
            "billingPostalCode": "75244-1111",
            "billingState": "TX",
        },
        "currencyShortName": "USD",
        "hardware": [
            {
                "accountId": 1111111,
                "bareMetalInstanceFlag": 0,
                "domain": "testedit.com",
                "fullyQualifiedDomainName": "test.testedit.com",
                "hostname": "test",
                "globalIdentifier": "81434794-af69-44d5-bb97-1111111"
            }
        ],
        "location": "1441195",
        "locationObject": {
            "id": 1441195,
            "longName": "Dallas 10",
            "name": "dal10"
        },
        "packageId": 911,
        "paymentType": "ADD_TO_BALANCE",
        "postTaxRecurring": "0",
        "postTaxRecurringHourly": "0",
        "postTaxRecurringMonthly": "0",
        "postTaxSetup": "0",
        "preTaxRecurring": "0",
        "preTaxRecurringHourly": "0",
        "preTaxRecurringMonthly": "0",
        "preTaxSetup": "0",
        "prices": [
            {
                "hourlyRecurringFee": "0",
                "id": 209391,
                "recurringFee": "0",
                "categories": [
                    {
                        "categoryCode": "ram",
                        "id": 3,
                        "name": "RAM"
                    }
                ],
                "item": {
                    "capacity": "32",
                    "description": "32 GB RAM",
                    "id": 11291,
                    "keyName": "RAM_32_GB_DDR4_2133_ECC_NON_REG",
                    "units": "GB",
                }
            }
        ],
        "proratedInitialCharge": "0",
        "proratedOrderTotal": "0",
        "quantity": 1,
        "totalRecurringTax": "0",
        "useHourlyPricing": False
    },
    "orderId": 78332111,
    "placedOrder": {
        "accountId": 1111111,
        "id": 1234,
        "status": "PENDING_UPGRADE",
        "account": {
            "brandId": 2,
            "companyName": "SoftLayer Internal - Development Community",
            "id": 1234
        },
        "items": [
            {
                "categoryCode": "ram",
                "description": "32 GB RAM",
                "id": 824199364,
                "recurringFee": "0"
            }
        ],
        "userRecord": {
            "accountId": 1234,
            "firstName": "test",
            "id": 3333,
            "lastName": "test",
            "username": "sl1234-test"
        }
    }
}

rsc_placeOrder = {
    'orderDate': '2013-08-01 15:23:45',
    'orderId': 1234,
    'orderDetails': {
        'postTaxRecurring': '0.32',
    },
    'placedOrder': {
        'status': 'Great, thanks for asking',
        'locationObject': {
            'id': 1854895,
            'longName': 'Dallas 13',
            'name': 'dal13'
        },
        'name': 'test-capacity',
        'items': [
            {
                'description': 'B1.1x2 (1 Year ''Term)',
                'keyName': 'B1_1X2_1_YEAR_TERM',
                'categoryCode': 'guest_core',
            }
        ]
    }
}

rsi_placeOrder = {
    'orderId': 1234,
    'orderDetails': {
        'prices': [
            {
                'id': 4,
                'item': {
                    'id': 1,
                    'description': 'B1.1x2 (1 Year ''Term)',
                    'keyName': 'B1_1X2_1_YEAR_TERM',
                },
                'hourlyRecurringFee': 1.0,
                'recurringFee': 2.0
            }
        ]
    }
}

wmware_placeOrder = {
    "orderDate": "2021-06-02 15:23:47",
    "orderId": 123456,
    "prices": [
        {
            "id": 176535,
            "itemId": 8109,
            "categories": [
                {
                    "categoryCode": "software_license",
                    "id": 438,
                    "name": "Software License"
                }
            ],
            "item": {
                "capacity": "1",
                "description": "VMware vSAN Advanced Tier III 64 - 124 TB 6.x",
                "id": 8109,
                "keyName": "VMWARE_VSAN_ADVANCE_TIER_III_64_124_6_X",
                "softwareDescription": {
                    "id": 1795,
                },
                "thirdPartyPolicyAssignments": [
                    {
                        "id": 29263,
                        "policyName": "3rd Party Software Terms VMWare v4"
                    }
                ]
            }
        }
    ]}

vlan_placeOrder = {"orderDate": "2021-06-02 15:23:47",
                   "orderId": 123456,
                   "orderDetails": {
                       "orderContainers": [{
                           "name": "test"}]},
                   "prices": [{
                       "id": 2018,
                       "itemId": 1071,
                       "categories": [{
                           "categoryCode": "network_vlan",
                           "id": 113,
                           "name": "Network Vlan"}],
                       "item": {
                           "capacity": "0",
                           "description": "Public Network Vlan",
                           "id": 1071,
                           "keyName": "PUBLIC_NETWORK_VLAN"}}
                   ]}

ipsec_placeOrder = {
    "orderDate": "2022-07-14T16:09:08-06:00",
    "orderId": 123456,
    "placedOrder": {
        "items": [
            {
                "categoryCode": "network_tunnel",
                "description": "IPSEC - Standard",
                "id": 931479898,
                "itemId": 1092,
                "itemPriceId": "2048",

            }
        ]

    }
}



"""List virtual servers."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils

# pylint: disable=unnecessary-lambda

COLUMNS = [
    column_helper.Column('guid', ('globalIdentifier',)),
    column_helper.Column('primary_ip', ('primaryIpAddress',)),
    column_helper.Column('backend_ip', ('primaryBackendIpAddress',)),
    column_helper.Column('datacenter', ('datacenter', 'name')),
    column_helper.Column('action', lambda guest: formatting.active_txn(guest),
                         mask='activeTransaction[id,transactionStatus[name,friendlyName]]'),
    column_helper.Column('power_state', ('powerState', 'name')),
    column_helper.Column('created_by', lambda created_by:
                         utils.lookup(created_by, 'billingItem', 'orderItem', 'order', 'userRecord', 'username'),
                         mask='billingItem[id,orderItem[id,order[id,userRecord[username]]]]'),
    column_helper.Column('tags', lambda server: formatting.tags(server.get('tagReferences')),
                         mask="tagReferences.tag.name"),
    column_helper.Column(
        'createDate',
        lambda guest: utils.clean_time(guest.get('createDate'),
                                       in_format='%Y-%m-%dT%H:%M:%S', out_format='%Y-%m-%d %H:%M'), mask="createDate"),
]

DEFAULT_COLUMNS = [
    'id',
    'hostname',
    'domain',
    'deviceStatus.name',
    'datacenter',
    'primary_ip',
    'backend_ip',
    'createDate',
    'action',
]


@click.command(cls=SLCommand, short_help="List virtual servers.")
@click.option('--cpu', '-c', help='Number of CPU cores', type=click.INT)
@click.option('--domain', '-D', help='Domain portion of the FQDN')
@click.option('--datacenter', '-d', help='Datacenter shortname')
@click.option('--hostname', '-H', help='Host portion of the FQDN')
@click.option('--memory', '-m', help='Memory in mebibytes', type=click.INT)
@click.option('--network', '-n', help='Network port speed in Mbps')
@click.option('--hourly', is_flag=True, help='Show only hourly instances')
@click.option('--monthly', is_flag=True, help='Show only monthly instances')
@click.option('--tag', '-t', help='list of tags', multiple=True)
@click.option('--transient', help='Filter by transient instances', type=click.BOOL)
@click.option('--search', is_flag=False, flag_value="", default=None,
              help="Use the more flexible Search API to list instances. See `slcli search --types` for list " +
              "of searchable fields.")
@helpers.multi_option('--tag', help='Filter by tags')
@click.option('--sortby', default='hostname', show_default=True, help='Column to sort by')
@click.option('--columns',
              callback=column_helper.get_formatter(COLUMNS),
              help=f"Columns to display. [options: {', '.join(column.name for column in COLUMNS)}]",
              default=','.join(DEFAULT_COLUMNS),
              show_default=True)
@click.option('--limit', '-l', default=100, show_default=True,
              help='How many results to get in one api call, default is 100')
@environment.pass_env
def cli(env, sortby, cpu, domain, datacenter, hostname, memory, network,
        hourly, monthly, tag, columns, limit, transient, search):
    """List virtual servers."""

    guests = []
    if search is not None:
        object_mask = f"mask[resource(SoftLayer_Virtual_Guest)[{columns.mask()}]]"
        search_manager = SoftLayer.SearchManager(env.client)
        guests = search_manager.search_instances(hostname=hostname, domain=domain, datacenter=datacenter,
                                                 tags=tag, search_string=search, mask=object_mask)
    else:
        vsi = SoftLayer.VSManager(env.client)
        guests = vsi.list_instances(hourly=hourly, monthly=monthly, hostname=hostname, domain=domain,
                                    cpus=cpu, memory=memory, datacenter=datacenter, nic_speed=network,
                                    transient=transient, tags=tag, mask=columns.mask(), limit=limit)

    table = formatting.Table(columns.columns)
    table.sortby = sortby

    for guest in guests:
        table.add_row([value or formatting.blank()
                       for value in columns.row(guest)])

    env.fout(table)



"""List package categories."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import ordering

COLUMNS = ['name', 'categoryCode', 'isRequired']


@click.command(cls=SLCommand)
@click.argument('package_keyname')
@click.option('--required',
              is_flag=True,
              help="List only the required categories for the package")
@environment.pass_env
def cli(env, package_keyname, required):
    """List the categories of a package.

    ::

        # List the categories of Bare Metal servers
        slcli order category-list BARE_METAL_SERVER

        # List the required categories for Bare Metal servers
        slcli order category-list BARE_METAL_SERVER --required

    """
    client = env.client
    manager = ordering.OrderingManager(client)
    table = formatting.Table(COLUMNS)

    categories = manager.list_categories(package_keyname)

    if required:
        categories = [cat for cat in categories if cat['isRequired']]

    for cat in categories:
        table.add_row([
            cat['itemCategory']['name'],
            cat['itemCategory']['categoryCode'],
            'Y' if cat['isRequired'] else 'N'
        ])

    env.fout(table)



getAllObjects = [
    {
        'backendRouterId': 117917,
        'backendRouterName': 'bcr01a.ams01',
        'datacenterId': 265592,
        'datacenterLongName': 'Amsterdam 1',
        'datacenterName': 'ams01',
        'frontendRouterId': 117960,
        'frontendRouterName': 'fcr01a.ams01',
        'name': 'ams01.pod01'
    },
    {
        'backendRouterId': 1115295,
        'backendRouterName': 'bcr01a.wdc07',
        'datacenterId': 2017603,
        'datacenterLongName': 'Washington 7',
        'datacenterName': 'wdc07',
        'frontendRouterId': 1114993,
        'frontendRouterName': 'fcr01a.wdc07',
        'name': 'wdc07.pod01'
    },
    {
        'backendRouterId': 1234567,
        'backendRouterName': 'bcr01a.wdc07',
        'datacenterId': 2017603,
        'datacenterLongName': 'Washington 7',
        'datacenterName': 'wdc07',
        'frontendRouterId': 258741369,
        'frontendRouterName': 'fcr01a.wdc07',
        'name': 'TEST00.pod2'
    }
]



"""Unassigns a global IP from a target."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Unroutes IDENTIFIER

    IDENTIFIER should be either the Global IP address, or the SoftLayer_Network_Subnet_IpAddress_Global id
    Example::

        slcli globalip unassign 123456

        slcli globalip unassign 123.43.22.11
"""

    mgr = SoftLayer.NetworkManager(env.client)
    global_ip_id = helpers.resolve_id(mgr.resolve_global_ip_ids, identifier, name='global ip')

    # Find Global IPs SoftLayer_Network_Subnet::id
    mask = "mask[id,ipAddress[subnetId]]"
    subnet = env.client.call('SoftLayer_Network_Subnet_IpAddress_Global', 'getObject', id=global_ip_id, mask=mask)
    subnet_id = subnet.get('ipAddress', {}).get('subnetId')
    mgr.clear_route(subnet_id)



"""Sync DNS records."""
# :license: MIT, see LICENSE for more details.
# pylint: disable=duplicate-code

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, epilog="""If you don't specify any
arguments, it will attempt to update both the A and PTR records. If you don't
want to update both records, you may use the -a or --ptr arguments to limit
the records updated.""")
@click.argument('identifier')
@click.option('--a-record', '-a', is_flag=True, help="Sync the A record for the host")
@click.option('--aaaa-record', is_flag=True, help="Sync the AAAA record for the host")
@click.option('--ptr', is_flag=True, help="Sync the PTR record for the host")
@click.option('--ttl', default=7200, show_default=True, type=click.INT,
              help="Sets the TTL for the A and/or PTR records")
@environment.pass_env
def cli(env, identifier, a_record, aaaa_record, ptr, ttl):
    """Sync DNS records."""

    mask = """mask[id, globalIdentifier, fullyQualifiedDomainName, hostname, domain,
              primaryBackendIpAddress,primaryIpAddress,
              primaryNetworkComponent[id,primaryIpAddress,primaryVersion6IpAddressRecord[ipAddress]]]"""
    dns = SoftLayer.DNSManager(env.client)
    server = SoftLayer.HardwareManager(env.client)

    server_id = helpers.resolve_id(server.resolve_ids, identifier, 'VS')
    instance = server.get_hardware(server_id, mask=mask)
    zone_id = helpers.resolve_id(dns.resolve_ids, instance['domain'], name='zone')

    if not instance['primaryIpAddress']:
        raise exceptions.CLIAbort('No primary IP address associated with this hardware')

    go_for_it = env.skip_confirmations or formatting.confirm(
        "Attempt to update DNS records for %s" % instance['fullyQualifiedDomainName'])

    if not go_for_it:
        raise exceptions.CLIAbort("Aborting DNS sync")

    # both will be true only if no options are passed in, basically.
    both = (not ptr) and (not a_record) and (not aaaa_record)

    if both or a_record:
        dns.sync_host_record(zone_id, instance['hostname'], instance['primaryIpAddress'], 'a', ttl)

    if both or ptr:
        # getReverseDomainRecords returns a list of 1 element, so just get the top.
        ptr_domains = env.client['Hardware_Server'].getReverseDomainRecords(id=instance['id']).pop()
        dns.sync_ptr_record(ptr_domains, instance['primaryIpAddress'], instance['fullyQualifiedDomainName'], ttl)

    if aaaa_record:
        try:
            # done this way to stay within 80 character lines
            ipv6 = instance['primaryNetworkComponent']['primaryVersion6IpAddressRecord']['ipAddress']
            dns.sync_host_record(zone_id, instance['hostname'], ipv6, 'aaaa', ttl)
        except KeyError as ex:
            message = f"{instance['fullyQualifiedDomainName']} does not have an ipv6 address"
            raise exceptions.CLIAbort(message) from ex



"""Display details for a specified cloud object storage."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('object_id')
@environment.pass_env
def cli(env, object_id):
    """Display details for a cloud object storage."""

    block_manager = SoftLayer.BlockStorageManager(env.client)

    cloud = block_manager.get_volume_details(object_id)
    bucket = block_manager.get_buckets(object_id)

    table = formatting.KeyValueTable(['Name', 'Value'])
    table.align['Name'] = 'r'
    table.align['Value'] = 'l'

    table.add_row(['Id', cloud.get('id')])
    table.add_row(['Username', cloud.get('username')])
    table.add_row(['Name Service Resource', cloud['serviceResource']['name']])
    table.add_row(['Type Service Resource', cloud['serviceResource']['type']['type']])
    table.add_row(['Datacenter', cloud['serviceResource']['datacenter']['name']])
    table.add_row(['Storage type', cloud['storageType']['keyName']])
    table.add_row(['Bytes Used', formatting.b_to_gb(bucket[0]['bytesUsed'])])
    table.add_row(['Bucket name', bucket[0]['name']])

    env.fout(table)



"""Manage security group rules."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting

COLUMNS = ['id',
           'remoteIp',
           'remoteGroupId',
           'direction',
           'ethertype',
           'portRangeMin',
           'portRangeMax',
           'protocol',
           'createDate',
           'modifyDate']

REQUEST_BOOL_COLUMNS = ['requestId', 'response']
REQUEST_RULES_COLUMNS = ['requestId', 'rules']


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('securitygroup_id')
@click.option('--sortby',
              help='Column to sort by',
              type=click.Choice(COLUMNS))
@environment.pass_env
def rule_list(env, securitygroup_id, sortby):
    """List security group rules."""

    mgr = SoftLayer.NetworkManager(env.client)

    table = formatting.Table(COLUMNS)
    table.sortby = sortby

    rules = mgr.list_securitygroup_rules(securitygroup_id)
    for rule in rules:
        port_min = rule.get('portRangeMin')
        port_max = rule.get('portRangeMax')
        if port_min is None:
            port_min = formatting.blank()
        if port_max is None:
            port_max = formatting.blank()

        table.add_row([
            rule['id'],
            rule.get('remoteIp') or formatting.blank(),
            rule.get('remoteGroupId') or formatting.blank(),
            rule['direction'],
            rule.get('ethertype') or formatting.blank(),
            port_min,
            port_max,
            rule.get('protocol') or formatting.blank(),
            rule.get('createDate') or formatting.blank(),
            rule.get('modifyDate') or formatting.blank()
        ])

    env.fout(table)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('securitygroup_id')
@click.option('--remote-ip', '-r',
              help='The remote IP/CIDR to enforce')
@click.option('--remote-group', '-s', type=click.INT,
              help='The ID of the remote security group to enforce')
@click.option('--direction', '-d',
              help=('The direction of traffic to enforce '
                    '(ingress, egress)'))
@click.option('--ethertype', '-e',
              help='The ethertype (IPv4 or IPv6) to enforce')
@click.option('--port-max', '-M', type=click.INT,
              help=('The upper port bound to enforce. When the protocol is ICMP, '
                    'this specifies the ICMP code to permit'))
@click.option('--port-min', '-m', type=click.INT,
              help=('The lower port bound to enforce. When the protocol is ICMP, '
                    'this specifies the ICMP type to permit'))
@click.option('--protocol', '-p',
              help='The protocol (icmp, tcp, udp) to enforce')
@environment.pass_env
def add(env, securitygroup_id, remote_ip, remote_group,
        direction, ethertype, port_max, port_min, protocol):
    """Add a security group rule to a security group.

    \b
    Examples:
        # Add an SSH rule (TCP port 22) to a security group
        slcli sg rule-add 384727 \\
            --direction ingress \\
            --protocol tcp \\
            --port-min 22 \\
            --port-max 22

    \b
        # Add a ping rule (ICMP type 8 code 0) to a security group
        slcli sg rule-add 384727 \\
            --direction ingress \\
            --protocol icmp \\
            --port-min 8 \\
            --port-max 0
    """
    mgr = SoftLayer.NetworkManager(env.client)

    ret = mgr.add_securitygroup_rule(securitygroup_id, remote_ip, remote_group,
                                     direction, ethertype, port_max,
                                     port_min, protocol)

    if not ret:
        raise exceptions.CLIAbort("Failed to add security group rule")

    table = formatting.Table(REQUEST_RULES_COLUMNS)
    table.add_row([ret['requestId'], str(ret['rules'])])

    env.fout(table)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('securitygroup_id')
@click.argument('rule_id')
@click.option('--remote-ip', '-r',
              help='The remote IP/CIDR to enforce')
@click.option('--remote-group', '-s',
              help='The ID of the remote security group to enforce')
@click.option('--direction', '-d',
              help='The direction of traffic to enforce')
@click.option('--ethertype', '-e',
              help='The ethertype (IPv4 or IPv6) to enforce')
@click.option('--port-max', '-M',
              help='The upper port bound to enforce')
@click.option('--port-min', '-m',
              help='The lower port bound to enforce')
@click.option('--protocol', '-p',
              help='The protocol (icmp, tcp, udp) to enforce')
@environment.pass_env
def edit(env, securitygroup_id, rule_id, remote_ip, remote_group,
         direction, ethertype, port_max, port_min, protocol):
    """Edit a security group rule in a security group."""
    mgr = SoftLayer.NetworkManager(env.client)

    data = {}
    if remote_ip:
        data['remote_ip'] = remote_ip
    if remote_group:
        data['remote_group'] = remote_group
    if direction:
        data['direction'] = direction
    if ethertype:
        data['ethertype'] = ethertype
    if port_max is not None:
        data['port_max'] = port_max
    if port_min is not None:
        data['port_min'] = port_min
    if protocol:
        data['protocol'] = protocol

    ret = mgr.edit_securitygroup_rule(securitygroup_id, rule_id, **data)

    if not ret:
        raise exceptions.CLIAbort("Failed to edit security group rule")

    table = formatting.Table(REQUEST_BOOL_COLUMNS)
    table.add_row([ret['requestId']])

    env.fout(table)


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('securitygroup_id')
@click.argument('rule_id')
@environment.pass_env
def remove(env, securitygroup_id, rule_id):
    """Remove a rule from a security group."""
    mgr = SoftLayer.NetworkManager(env.client)

    ret = mgr.remove_securitygroup_rule(securitygroup_id, rule_id)

    if not ret:
        raise exceptions.CLIAbort("Failed to remove security group rule")

    table = formatting.Table(REQUEST_BOOL_COLUMNS)
    table.add_row([ret['requestId']])

    env.fout(table)



"""Edit details of an Delivery email account."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.managers.email import EmailManager


@click.command(cls=SLCommand)
@click.argument('identifier')
@click.option('--username', help="Sets username for this account")
@click.option('--email', help="Sets the contact email for this account")
@click.option('--password',
              help="Password must be between 8 and 20 characters "
                   "and must contain one letter and one number.")
@environment.pass_env
def cli(env, identifier, username, email, password):
    """Edit details of an email delivery account."""
    email_manager = EmailManager(env.client)

    data = {}
    update = False
    if email:
        if email_manager.update_email(identifier, email):
            update = True
        else:
            raise exceptions.CLIAbort("Failed to Edit emailAddress account")
    if username:
        data['username'] = username
    if password:
        data['password'] = password
    if len(data) != 0:
        if email_manager.editObject(identifier, **data):
            update = True
        else:
            raise exceptions.CLIAbort("Failed to Edit email account")

    if update:
        env.fout('Updated Successfully')



"""Get details for a hardware device."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils


# pylint: disable=R0915


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--passwords', is_flag=True, help='Show passwords (check over your shoulder!)')
@click.option('--price', is_flag=True, help='Show associated prices')
@click.option('--components', is_flag=True, default=False, help='Show associated hardware components')
@environment.pass_env
def cli(env, identifier, passwords, price, components):
    """Get details for a hardware device."""

    hardware = SoftLayer.HardwareManager(env.client)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    hardware_id = helpers.resolve_id(hardware.resolve_ids, identifier, 'hardware')
    result = hardware.get_hardware_fast(hardware_id)
    result = utils.NestedDict(result)
    hard_drives = hardware.get_hard_drives(hardware_id)

    operating_system = utils.lookup(result, 'operatingSystem', 'softwareLicense', 'softwareDescription') or {}
    memory = formatting.gb(result.get('memoryCapacity', 0))
    owner = None
    if utils.lookup(result, 'billingItem'):
        owner = utils.lookup(result, 'billingItem', 'orderItem', 'order', 'userRecord', 'username')

    table_hard_drives = formatting.Table(['Name', 'Capacity', 'Serial #'])
    for drives in hard_drives:
        name = drives['hardwareComponentModel']['manufacturer'] + " " + drives['hardwareComponentModel']['name']
        capacity = str(drives['hardwareComponentModel']['hardwareGenericComponentModel']['capacity']) + " " + str(
            drives['hardwareComponentModel']['hardwareGenericComponentModel']['units'])
        serial = drives['serialNumber']

        table_hard_drives.add_row([name, capacity, serial])

    table.add_row(['id', result['id']])
    table.add_row(['guid', result['globalIdentifier'] or formatting.blank()])
    table.add_row(['hostname', result['hostname']])
    table.add_row(['domain', result['domain']])
    table.add_row(['fqdn', result['fullyQualifiedDomainName']])
    table.add_row(['status', result['hardwareStatus']['status']])
    table.add_row(['datacenter', result['datacenter']['name'] or formatting.blank()])
    table.add_row(['cores', result['processorPhysicalCoreAmount']])
    table.add_row(['memory', memory])
    table.add_row(['drives', table_hard_drives])
    table.add_row(['public_ip', result['primaryIpAddress'] or formatting.blank()])
    table.add_row(['private_ip', result['primaryBackendIpAddress'] or formatting.blank()])
    table.add_row(['ipmi_ip', result['networkManagementIpAddress'] or formatting.blank()])
    table.add_row(['os', operating_system.get('name') or formatting.blank()])
    table.add_row(['os_version', operating_system.get('version') or formatting.blank()])
    table.add_row(['created', result['provisionDate'] or formatting.blank()])
    table.add_row(['owner', owner or formatting.blank()])

    last_transaction = f"{utils.lookup(result, 'lastTransaction', 'transactionGroup', 'name')} \
        ({utils.clean_time(utils.lookup(result, 'lastTransaction', 'modifyDate'))})"

    table.add_row(['last_transaction', last_transaction])
    table.add_row(['billing', 'Hourly' if result['hourlyBillingFlag'] else 'Monthly'])

    vlan_table = formatting.Table(['Network', 'Number', 'Id', 'Name', 'Type'])
    for vlan in result['networkVlans']:
        vlan_table.add_row([
            vlan.get('networkSpace'),
            vlan.get('vlanNumber'),
            vlan['id'],
            vlan['fullyQualifiedName'],
            'Primary'
        ])

    # Shows any VLANS trunked/tagged on this server
    for component in result.get('networkComponents', []):
        # These are the Primary network components
        if component.get('primaryIpAddress', False):
            uplink = component.get('uplinkComponent', {})
            for trunk in uplink.get('networkVlanTrunks', []):
                trunk_vlan = trunk.get('networkVlan')
                vlan_table.add_row([
                    trunk_vlan.get('networkSpace'),
                    trunk_vlan.get('vlanNumber'),
                    trunk_vlan.get('id'),
                    trunk_vlan.get('fullyQualifiedName'),
                    'Trunked'
                ])

    table.add_row(['vlans', vlan_table])

    bandwidth = hardware.get_bandwidth_allocation(hardware_id)
    bw_table = _bw_table(bandwidth)
    table.add_row(['Bandwidth', bw_table])
    system_table = _system_table(result['activeComponents'])
    table.add_row(['System_data', system_table])

    if result.get('notes'):
        table.add_row(['notes', result['notes']])

    if price:
        total_price = utils.lookup(result, 'billingItem', 'nextInvoiceTotalRecurringAmount') or 0

        price_table = formatting.Table(['Item', 'CategoryCode', 'Recurring Price'])
        price_table.align['Item'] = 'l'

        price_table.add_row(['Total', '-', total_price])

        for item in utils.lookup(result, 'billingItem', 'nextInvoiceChildren') or []:
            price_table.add_row([item['description'], item['categoryCode'], item['nextInvoiceTotalRecurringAmount']])

        table.add_row(['prices', price_table])

    if passwords:
        pass_table = formatting.Table(['username', 'password'])
        for item in result['operatingSystem']['passwords']:
            pass_table.add_row([item['username'], item['password']])
        table.add_row(['users', pass_table])

        pass_table = formatting.Table(['ipmi_username', 'password'])
        for item in result['remoteManagementAccounts']:
            pass_table.add_row([item['username'], item['password']])
        table.add_row(['remote users', pass_table])

    if components:
        components = hardware.get_components(identifier)
        components_table = formatting.Table(['name', 'Firmware version', 'Firmware build date', 'Type'])
        components_table.align['date'] = 'l'
        component_ids = []
        for hw_component in components:
            if hw_component['id'] not in component_ids:
                firmware = hw_component['hardwareComponentModel']['firmwares'][0]
                components_table.add_row([utils.lookup(hw_component, 'hardwareComponentModel', 'longDescription'),
                                          utils.lookup(firmware, 'version'),
                                          utils.clean_time(utils.lookup(firmware, 'createDate')),
                                          utils.lookup(hw_component, 'hardwareComponentModel',
                                                       'hardwareGenericComponentModel', 'hardwareComponentType',
                                                       'keyName')])
                component_ids.append(hw_component['id'])

        table.add_row(['components', components_table])

    table.add_row(['tags', formatting.tags(result['tagReferences'])])

    env.fout(table)


def _bw_table(bw_data):
    """Generates a bandwidth usage table"""
    table = formatting.Table(['Type', 'In GB', 'Out GB', 'Allotment'])
    for bw_point in bw_data.get('usage'):
        bw_type = 'Private'
        allotment = 'N/A'
        if bw_point['type']['alias'] == 'PUBLIC_SERVER_BW':
            bw_type = 'Public'
            if not bw_data.get('allotment'):
                allotment = '-'
            else:
                allotment = utils.lookup(bw_data, 'allotment', 'amount')

        table.add_row([bw_type, bw_point['amountIn'], bw_point['amountOut'], allotment])
    return table


def _system_table(system_data):
    table = formatting.Table(['Type', 'Name'])
    table.align['Type'] = 'r'
    table.align['Name'] = 'l'
    table.sortby = 'Type'
    for system in system_data:
        c_type = utils.lookup(system, 'hardwareComponentModel', 'hardwareGenericComponentModel',
                              'hardwareComponentType', 'keyName')
        c_name = utils.lookup(system, 'hardwareComponentModel', 'longDescription')
        table.add_row([c_type, c_name])
    return table



"""Virtual Servers."""
# :license: MIT, see LICENSE for more details.

import re

import click

MEMORY_RE = re.compile(r"^(?P<amount>[0-9]+)(?P<unit>g|gb|m|mb)?$")


class MemoryType(click.ParamType):
    """Memory type."""
    name = 'integer'

    def convert(self, value, param, ctx):  # pylint: disable=inconsistent-return-statements
        """Validate memory argument. Returns the memory value in megabytes."""
        matches = MEMORY_RE.match(value.lower())
        if matches is None:
            self.fail('%s is not a valid value for memory amount' % value, param, ctx)
        amount_str, unit = matches.groups()
        amount = int(amount_str)
        if unit in [None, 'm', 'mb']:
            # Assume the user intends gigabytes if they specify a number < 1024
            if amount < 1024:
                return amount * 1024
            else:
                if amount % 1024 != 0:
                    self.fail('%s is not an integer that is divisable by 1024' % value, param, ctx)
                return amount
        elif unit in ['g', 'gb']:
            return amount * 1024


MEM_TYPE = MemoryType()



"""
    SoftLayer.CLI.core
    ~~~~~~~~~~~~~~~~~~
    Core for the SoftLayer CLI

    :license: MIT, see LICENSE for more details.
"""
import logging
import os
import sys
import time
import traceback

import click
import requests
from rich.markup import escape

import SoftLayer
from SoftLayer.CLI.command import CommandLoader
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer import consts

# pylint: disable=too-many-public-methods, broad-except, unused-argument, invalid-name
# pylint: disable=redefined-builtin, super-init-not-called, arguments-differ

START_TIME = time.time()
DEBUG_LOGGING_MAP = {
    0: logging.CRITICAL,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG
}

PROG_NAME = "slcli (SoftLayer Command-line)"
VALID_FORMATS = ['table', 'raw', 'json', 'jsonraw', 'csv']
DEFAULT_FORMAT = 'raw'

if sys.stdout.isatty():
    DEFAULT_FORMAT = 'table'


def get_latest_version():
    """Gets the latest version of the Softlayer library."""
    try:
        result = requests.get('https://pypi.org/pypi/SoftLayer/json',  timeout=60)
        json_result = result.json()
        latest = f"v{json_result['info']['version']}"
    except Exception:
        latest = "Unable to get version from pypi."
    return latest


CONTEXT_SETTINGS = {
    "help_option_names": ['--help', '-h'],
    "auto_envvar_prefix": 'SLCLI',
    "max_content_width": 999
}


def get_version_message(ctx, param, value):
    """Gets current and latest release versions message."""
    if not value or ctx.resilient_parsing:
        return
    current = SoftLayer.consts.VERSION
    latest = get_latest_version()
    click.secho(f"Current: {PROG_NAME} {current}\nLatest:  {PROG_NAME} {latest}")
    ctx.exit()


@click.group(help="SoftLayer Command-line Client",
             epilog="""To use most commands your SoftLayer username and api_key need to be configured.
The easiest way to do that is to use: 'slcli setup'""",
             cls=CommandLoader,
             context_settings=CONTEXT_SETTINGS)
@click.option('--format', default=DEFAULT_FORMAT, show_default=True,
              help="Output format",
              type=click.Choice(VALID_FORMATS))
@click.option('-C', '--config', required=False, show_default=True,
              default=click.get_app_dir('softlayer', force_posix=True),
              help="Config file location",
              type=click.Path(resolve_path=True))
@click.option('--verbose', '-v',
              help="Sets the debug noise level, specify multiple times for more verbosity.",
              type=click.IntRange(0, 3, clamp=True),
              count=True)
@click.option('--proxy', required=False,
              help="HTTPS or HTTP proxy to be use to make API calls")
@click.option('--really / --not-really', '-y', is_flag=True, required=False,
              help="Confirm all prompt actions")
@click.option('--demo / --no-demo', is_flag=True, required=False,
              help="Use demo data instead of actually making API calls")
@click.option('--version', is_flag=True, expose_value=False, is_eager=True, callback=get_version_message,
              help="Show version information.",  allow_from_autoenv=False,)
@click.option('--account', '-a', help="Account Id, only needed for some API calls.")
@click.option('--internal', '-i', is_flag=True, required=False,
              help="Use the Employee Client instead of the Customer Client.")
@environment.pass_env
def cli(env,
        format='table',
        config=None,
        verbose=0,
        proxy=None,
        really=False,
        demo=False,
        account=None,
        internal=False,
        **kwargs):
    """Main click CLI entry-point."""

    # Populate environment with client and set it as the context object
    env.skip_confirmations = really
    env.config_file = config
    env.format = format
    env.set_env_theme(config_file=config)
    if internal:
        env.ensure_emp_client(config_file=config, is_demo=demo, proxy=proxy)
    else:
        env.ensure_client(config_file=config, is_demo=demo, proxy=proxy)
    env.vars['_start'] = time.time()
    logger = logging.getLogger()

    if demo is False:
        logger.addHandler(logging.StreamHandler())
    else:
        # This section is for running CLI tests.
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logger.addHandler(logging.NullHandler())

    logger.setLevel(DEBUG_LOGGING_MAP.get(verbose, logging.DEBUG))
    env.vars['_timings'] = SoftLayer.DebugTransport(env.client.transport)
    env.vars['verbose'] = verbose
    env.client.transport = env.vars['_timings']
    env.client.account_id = account


@cli.result_callback()
@environment.pass_env
def output_diagnostics(env, result, verbose=0, **kwargs):
    """Output diagnostic information."""

    if verbose > 0:
        diagnostic_table = formatting.Table(['name', 'value'])
        diagnostic_table.add_row(['execution_time', '%fs' % (time.time() - START_TIME)])

        api_call_value = formatting.Table(['API Calls'], title=None, align="left")
        for call in env.client.transport.get_last_calls():
            api_call_value.add_row(["%s::%s (%fs)" % (call.service, call.method, call.end_time - call.start_time)])

        diagnostic_table.add_row(['api_calls', api_call_value])
        diagnostic_table.add_row(['version', consts.USER_AGENT])
        diagnostic_table.add_row(['python_version', sys.version])
        diagnostic_table.add_row(['library_location', os.path.dirname(SoftLayer.__file__)])

        env.err(env.fmt(diagnostic_table))

    if verbose > 1:
        for call in env.client.transport.get_last_calls():
            call_table = formatting.Table(['', f'{call.service}::{call.method}'], align="left")
            nice_mask = ''
            if call.mask is not None:
                nice_mask = call.mask
            call_table.add_row(['id', call.identifier])
            #  Need to escape this so Rich doesn't think the mask is a tag and hide it.
            call_table.add_row(['mask', escape(nice_mask)])
            call_table.add_row(['filter', call.filter])
            call_table.add_row(['limit', call.limit])
            call_table.add_row(['offset', call.offset])
            env.err(env.fmt(call_table))

    if verbose > 2:
        for call in env.client.transport.get_last_calls():
            env.err(env.client.transport.print_reproduceable(call))


def main(reraise_exceptions=False, **kwargs):
    """Main program. Catches several common errors and displays them nicely."""
    exit_status = 0

    try:
        cli.main(**kwargs)
    except SoftLayer.SoftLayerAPIError as ex:
        if 'invalid api token' in ex.faultString.lower():
            print("Authentication Failed: To update your credentials, use 'slcli config setup'")
            exit_status = 1
        else:
            print(str(ex))
            exit_status = 1
    except SoftLayer.SoftLayerError as ex:
        print(str(ex))
        exit_status = 1
    except exceptions.CLIAbort as ex:
        print(str(ex.message))
        exit_status = ex.code
    except Exception:
        if reraise_exceptions:
            raise

        print("An unexpected error has occured:")
        print(str(traceback.format_exc()))
        print("Feel free to report this error as it is likely a bug:")
        print("    https://github.com/softlayer/softlayer-python/issues")
        print("The following snippet should be able to reproduce the error")
        exit_status = 1

    sys.exit(exit_status)


if __name__ == '__main__':
    main()



getObject = {
    'id': 1000,
    'globalIdentifier': '1a2b3c-1701',
    'datacenter': {'id': 50, 'name': 'TEST00', 'longName': 'test 00',
                   'description': 'Test Data Center'},
    'billingItem': {
        'id': 6327,
        'recurringFee': 1.54,
        'package': {
            'id': 911
        },
        'nextInvoiceTotalRecurringAmount': 16.08,
        'children': [
            {'description': 'test', 'nextInvoiceTotalRecurringAmount': 1},
        ],
        'nextInvoiceChildren': [
            {'description': 'test', 'nextInvoiceTotalRecurringAmount': 1, 'categoryCode': 'disk1'},
            {'description': 'test2', 'nextInvoiceTotalRecurringAmount': 2, 'categoryCode': 'disk3'}
        ],
        'orderItem': {
            'order': {
                'userRecord': {
                    'username': 'chechu',
                }
            }
        }
    },
    'primaryIpAddress': '172.16.1.100',
    'hostname': 'hardware-test1',
    'domain': 'test.sftlyr.ws',
    'bareMetalInstanceFlag': True,
    'fullyQualifiedDomainName': 'hardware-test1.test.sftlyr.ws',
    'processorPhysicalCoreAmount': 2,
    'memoryCapacity': 2,
    'primaryBackendIpAddress': '10.1.0.2',
    'networkManagementIpAddress': '10.1.0.3',
    'hardwareStatus': {'status': 'ACTIVE'},
    'primaryNetworkComponent': {'maxSpeed': 10, 'speed': 10},
    'provisionDate': '2013-08-01 15:23:45',
    'notes': 'These are test notes.',
    'operatingSystem': {
        'softwareLicense': {
            'softwareDescription': {
                'referenceCode': 'UBUNTU_12_64',
                'name': 'Ubuntu',
                'version': 'Ubuntu 12.04 LTS',
            }
        },
        'passwords': [
            {'username': 'root', 'password': 'abc123'}
        ],
    },
    'remoteManagementAccounts': [
        {'username': 'root', 'password': 'abc123'}
    ],
    'networkVlans': [
        {
            'networkSpace': 'PRIVATE',
            'vlanNumber': 1800,
            'id': 9653,
            'fullyQualifiedName': 'dal10.bcr03.14752',
            'primarySubnets': [{
                'netmask': ''
            }]
        },
        {
            'networkSpace': 'PUBLIC',
            'vlanNumber': 3672,
            'id': 19082,
            'fullyQualifiedName': 'dal10.test03.123',
            'primarySubnets': [{
                'netmask': ''
            }]
        },
    ],
    'tagReferences': [
        {'tag': {'name': 'test_tag'}}
    ],
    'activeTransaction': {
        'transactionStatus': {
            'name': 'TXN_NAME',
            'friendlyName': 'Friendly Transaction Name',
            'id': 6660
        }
    },
    'networkMonitors': [
        {
            'hardwareId': 3123796,
            'hostId': 3123796,
            'id': 19016454,
            'ipAddress': '169.53.167.199',
            'queryTypeId': 1,
            'responseActionId': 2,
            'status': 'ON',
            'waitCycles': 0,
            'lastResult': {
                'finishTime': '2022-03-10T08:31:40-06:00',
                'responseStatus': 2,
                'responseTime': 159.15,
            },
            'queryType': {
                'description': 'Test ping to address',
                'id': 1,
                'monitorLevel': 0,
                'name': 'SERVICE PING'
            },
            'responseAction': {
                'actionDescription': 'Notify Users',
                'id': 2,
                'level': 0
            }
        }
    ]
}

editObject = True
setTags = True
setPrivateNetworkInterfaceSpeed = True
setPublicNetworkInterfaceSpeed = True
toggleManagementInterface = True
powerOff = True
powerOn = True
powerCycle = True
rebootSoft = True
rebootDefault = True
rebootHard = True
createFirmwareUpdateTransaction = True
createFirmwareReflashTransaction = True
setUserMetadata = ['meta']
reloadOperatingSystem = 'OK'
getReverseDomainRecords = [
    {'resourceRecords': [{'data': '2.0.1.10.in-addr.arpa'}]}]
bootToRescueLayer = True

getNetworkComponents = [
    {'maxSpeed': 100},
    {
        'maxSpeed': 1000,
        'networkComponentGroup': {
            'groupTypeId': 2,
            'networkComponents': [{'maxSpeed': 1000}, {'maxSpeed': 1000}]
        },
        'primaryIpAddress': '192.168.1.1',
        'id': 998877,
        'uplinkComponent': {}
    },
    {
        'maxSpeed': 1000,
        'networkComponentGroup': {
            'groupTypeId': 2,
            'networkComponents': [{'maxSpeed': 1000}, {'maxSpeed': 1000}]
        },
        'id': 665544,
        'uplinkComponent': {}
    },
    {
        'maxSpeed': 1000,
        'networkComponentGroup': {
            'groupTypeId': 2,
            'networkComponents': [{'maxSpeed': 1000}, {'maxSpeed': 1000}]
        },
        'id': 112233,
        'uplinkComponent': {}
    },
    {
        'maxSpeed': 1000,
        'networkComponentGroup': {
            'groupTypeId': 2,
            'networkComponents': [{'maxSpeed': 1000}, {'maxSpeed': 1000}]
        },
        'primaryIpAddress': '10.0.0.1',
        'id': 123456,
        'uplinkComponent': {}
    }
]
# This splits out the network components into 2 sections so they are different enough for tests
getFrontendNetworkComponents = getNetworkComponents[:2]
getBackendNetworkComponents = getNetworkComponents[3:]

getBandwidthAllotmentDetail = {
    'allocationId': 25465663,
    'bandwidthAllotmentId': 138442,
    'effectiveDate': '2019-04-03T23:00:00-06:00',
    'endEffectiveDate': None,
    'id': 25888247,
    'serviceProviderId': 1,
    'allocation': {
        'amount': '250'
    }
}

getBillingCycleBandwidthUsage = [
    {
        'amountIn': '.448',
        'amountOut': '.52157',
        'type': {
            'alias': 'PUBLIC_SERVER_BW'
        }
    },
    {
        'amountIn': '.03842',
        'amountOut': '.01822',
        'type': {
            'alias': 'PRIVATE_SERVER_BW'
        }
    }
]

getMetricTrackingObjectId = 1000

getAttachedNetworkStorages = [
    {
        "accountId": 11111,
        "capacityGb": 20,
        "createDate": "2018-04-05T05:15:49-06:00",
        "id": 22222,
        "nasType": "NAS",
        "serviceProviderId": 1,
        "storageTypeId": "13",
        "username": "SL02SEV311111_11",
        "allowedHardware": [
            {
                "id": 12345,
                "datacenter": {
                    "id": 449506,
                    "longName": "Frankfurt 2",
                    "name": "fra02",
                    "statusId": 2
                }
            }
        ],
        "serviceResourceBackendIpAddress": "fsn-fra0201a-fz.service.softlayer.com",
        "serviceResourceName": "Storage Type 02 File Aggregate stfm-fra0201a"
    },
    {
        "accountId": 11111,
        "capacityGb": 12000,
        "createDate": "2018-01-28T04:57:30-06:00",
        "id": 3777111,
        "nasType": "ISCSI",
        "notes": "BlockStorage12T",
        "password": "",
        "serviceProviderId": 1,
        "storageTypeId": "7",
        "username": "SL02SEL32222-9",
        "allowedHardware": [
            {
                "id": 629222,
                "datacenter": {
                    "id": 449506,
                    "longName": "Frankfurt 2",
                    "name": "fra02",
                    "statusId": 2
                }
            }
        ],
        "serviceResourceBackendIpAddress": "10.31.95.152",
        "serviceResourceName": "Storage Type 02 Block Aggregate stbm-fra0201a"
    }
]

getAllowedHost = {
    "accountId": 11111,
    "credentialId": 22222,
    "id": 33333,
    "name": "iqn.2020-03.com.ibm:sl02su11111-v62941551",
    "resourceTableId": 6291111,
    "resourceTableName": "VIRTUAL_GUEST",
    "credential": {
        "accountId": "11111",
        "createDate": "2020-03-20T13:35:47-06:00",
        "id": 44444,
        "nasCredentialTypeId": 2,
        "password": "SjFDCpHrjskfj",
        "username": "SL02SU11111-V62941551"
    }
}

getHardDrives = [
    {
        "id": 11111,
        "serialNumber": "z1w4sdf",
        "serviceProviderId": 1,
        "hardwareComponentModel": {
            "capacity": "1000",
            "description": "SATAIII:2000:8300:Constellation",
            "id": 111,
            "manufacturer": "Seagate",
            "name": "Constellation ES",
            "hardwareGenericComponentModel": {
                "capacity": "1000",
                "units": "GB",
                "hardwareComponentType": {
                    "id": 1,
                    "keyName": "HARD_DRIVE",
                    "type": "Hard Drive",
                    "typeParentId": 5
                }
            }
        }
    }
]

getUpgradeItemPrices = [
    {
        "id": 21525,
        "recurringFee": "0",
        "categories": [
            {
                "categoryCode": "port_speed",
                "id": 26,
                "name": "Uplink Port Speeds",
            }
        ],
        "item": {
            "capacity": "10000",
            "description": "10 Gbps Redundant Public & Private Network Uplinks",
            "id": 4342,
            "keyName": "10_GBPS_REDUNDANT_PUBLIC_PRIVATE_NETWORK_UPLINKS"
        }
    },
    {
        "hourlyRecurringFee": ".247",
        "id": 209391,
        "recurringFee": "164",
        "categories": [
            {
                "categoryCode": "ram",
                "id": 3,
                "name": "RAM"
            }
        ],
        "item": {
            "capacity": "32",
            "description": "32 GB RAM",
            "id": 11291,
            "keyName": "RAM_32_GB_DDR4_2133_ECC_NON_REG"
        }
    },
    {
        "hourlyRecurringFee": ".068",
        "id": 22482,
        "recurringFee": "50",
        "categories": [
            {
                "categoryCode": "disk_controller",
                "id": 11,
                "name": "Disk Controller",
            }
        ],
        "item": {
            "capacity": "0",
            "description": "RAID",
            "id": 4478,
            "keyName": "DISK_CONTROLLER_RAID",
        }
    },
    {
        "id": 50357,
        "recurringFee": "0",
        "categories": [
            {
                "categoryCode": "bandwidth",
                "id": 10,
                "name": "Public Bandwidth",
            }
        ],
        "item": {
            "capacity": "500",
            "description": "500 GB Bandwidth Allotment",
            "id": 6177,
            "keyName": "BANDWIDTH_500_GB"
        }
    },
    {
        "hourlyRecurringFee": ".023",
        "id": 49759,
        "recurringFee": "15",
        "categories": [
            {
                "categoryCode": "disk2",
                "id": 6,
                "name": "Third Hard Drive"
            }
        ],
        "item": {
            "capacity": "1000",
            "description": "1.00 TB SATA",
            "id": 6159,
            "keyName": "HARD_DRIVE_1_00_TB_SATA_2",
        }
    },
    {
        "id": 49759,
        "recurringFee": "0",
        "categories": [
            {
                "categoryCode": "disk1",
                "id": 5,
                "name": "Second Hard Drive"
            }
        ],
        "item": {
            "capacity": "1000",
            "description": "1.00 TB SATA",
            "id": 6159,
            "keyName": "HARD_DRIVE_1_00_TB_SATA_2"
        }
    }
]

getComponents = [{
    "hardwareComponentModelId": 147,
    "hardwareId": 1234,
    "id": 369,
    "modifyDate": "2017-11-10T16:59:38-06:00",
    "serviceProviderId": 1,
    "hardwareComponentModel": {
        "name": "IMM2 - Onboard",
        "firmwares": [
            {
                "createDate": "2020-09-24T13:46:29-06:00",
                "version": "5.60"
            },
            {
                "createDate": "2019-10-14T16:51:12-06:00",
                "version": "5.10"
            }
        ]
    }
}]
getActiveComponents = getComponents
getActiveTransaction = getObject['activeTransaction']
getOperatingSystem = getObject['operatingSystem']
getSoftwareComponents = [
    {
        "hardwareId": 1907356,
        "id": 59003868,
        "manufacturerLicenseInstance": "",
        "softwareLicense": {
            "id": 20658,
            "softwareDescriptionId": 2888,
            "softwareDescription": {
                "controlPanel": 0,
                "id": 2888,
                "licenseTermValue": 0,
                "longDescription": "Juniper vSRX 1G 19.4R2-S3 Standard 19.4.2.3",
                "manufacturer": "Juniper",
                "name": "vSRX 1G 19.4R2-S3 Standard",
                "operatingSystem": 1,
                "referenceCode": "UBUNTU_18_64",
                "upgradeSoftwareDescriptionId": None,
                "upgradeSwDescId": None,
                "version": "19.4.2.3",
                "virtualLicense": 0,
                "virtualizationPlatform": 0,
                "requiredUser": "root"
            }
        }
    },
    {
        "hardwareId": 1907356,
        "id": 59003870,
        "manufacturerLicenseInstance": "",
        "softwareLicense": {
            "id": 147,
            "softwareDescriptionId": 148,
            "softwareDescription": {
                "controlPanel": 0,
                "id": 148,
                "licenseTermValue": None,
                "longDescription": "Passmark Suite Latest",
                "manufacturer": "Passmark",
                "name": "Passmark Suite",
                "operatingSystem": 0,
                "upgradeSoftwareDescriptionId": None,
                "upgradeSwDescId": None,
                "version": "Latest",
                "virtualLicense": 0,
                "virtualizationPlatform": 0
            }
        }
    }
]
getBillingItem = getObject['billingItem']
getTagReferences = getObject['tagReferences']
getNetworkVlans = getObject['networkVlans']
getRemoteManagementAccounts = getObject['remoteManagementAccounts']


# Setup for hardwareManager.clear_vlan related tests
getObjectVlanClear = {
    'backendNetworkComponent': getBackendNetworkComponents,
    'frontendNetworkComponent': getFrontendNetworkComponents
}
getObjectVlanClear['backendNetworkComponent'][1]['networkVlanTrunks'] = [{'id': 99}]
getObjectVlanClear['frontendNetworkComponent'][1]['networkVlanTrunks'] = [{'id': 11}]



"""Add a new load balancer protocol."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


# pylint: disable=unused-argument
def parse_proto(ctx, param, value):
    """Parses the frontend and backend cli options"""
    proto = {'protocol': 'HTTP', 'port': 80}
    splitout = value.split(':')
    if len(splitout) != 2:
        raise exceptions.ArgumentError("{}={} is not properly formatted.".format(param, value))
    proto['protocol'] = splitout[0]
    proto['port'] = int(splitout[1])
    return proto


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--frontend', '-f', required=True, default='HTTP:80', show_default=True, callback=parse_proto,
              help='PROTOCOL:PORT string for incoming internet connections.')
@click.option('--backend', '-b', required=True, default='HTTP:80', show_default=True, callback=parse_proto,
              help='PROTOCOL:PORT string for connecting to backend servers.')
@click.option('--method', '-m', help="Balancing Method.", default='ROUNDROBIN', show_default=True,
              type=click.Choice(['ROUNDROBIN', 'LEASTCONNECTION', 'WEIGHTED_RR']))
@click.option('--session', '-s', required=True,
              help="Session stickiness type. Valid values are SOURCE_IP or HTTP_COOKIE ")
@click.option('--max', help="Max Connections setting", type=int)
@environment.pass_env
def cli(env, identifier, **args):
    """Add a new load balancer protocol."""

    mgr = SoftLayer.LoadBalancerManager(env.client)

    uuid = mgr.get_lb_uuid(identifier)

    backend = args.get('backend')
    frontend = args.get('frontend')
    protocols = [
        {
            "backendPort": backend.get('port'),
            "backendProtocol": backend.get('protocol'),
            "frontendPort": frontend.get('port'),
            "frontendProtocol": frontend.get('protocol'),
            "loadBalancingMethod": args.get('method'),
            "sessionType": args.get('session'),
            "maxConn": args.get('max')
        }
    ]

    protocol = mgr.add_protocols(uuid, protocols)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Id', protocol.get('id')])
    table.add_row(['UUI', protocol.get('uuid')])
    table.add_row(['Address', protocol.get('address')])
    table.add_row(['Type', mgr.get_lb_type(protocol.get('type'))])
    table.add_row(['Description', protocol.get('description')])

    env.fout(table)



createObject = {'name': 'example.com'}
deleteObject = True
editObject = True
getZoneFileContents = 'lots of text'
getResourceRecords = [
    {'id': 1, 'ttl': 7200, 'data': 'd', 'host': 'a', 'type': 'cname'},
    {'id': 2, 'ttl': 900, 'data': '1', 'host': 'b', 'type': 'a'},
    {'id': 3, 'ttl': 900, 'data': 'x', 'host': 'c', 'type': 'ptr'},
    {'id': 4, 'ttl': 86400, 'data': 'b', 'host': 'd', 'type': 'txt'},
    {'id': 5, 'ttl': 86400, 'data': 'b', 'host': 'e', 'type': 'txt'},
    {'id': 6, 'ttl': 600, 'data': 'b', 'host': 'f', 'type': 'txt'},
]
getObject = {'id': 98765, 'name': 'test-example.com'}



"""
    SoftLayer.storage_utils
    ~~~~~~~~~~~~~~~~~~~~~~~
    Utility functions used by File and Block Storage Managers

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer import exceptions
from SoftLayer import utils

# pylint: disable=too-many-lines


ENDURANCE_TIERS = {
    0.25: 100,
    2: 200,
    4: 300,
    10: 1000,
}


def populate_host_templates(hardware_ids=None,
                            virtual_guest_ids=None,
                            ip_address_ids=None,
                            subnet_ids=None):
    """Returns a populated array with the IDs provided

    :param hardware_ids: A List of SoftLayer_Hardware ids
    :param virtual_guest_ids: A List of SoftLayer_Virtual_Guest ids
    :param ip_address_ids: A List of SoftLayer_Network_Subnet_IpAddress ids
    :param subnet_ids: A List of SoftLayer_Network_Subnet ids
    :return: array of objects formatted for allowAccessFromHostList
    """
    host_templates = []
    if hardware_ids is not None:
        for hardware_id in hardware_ids:
            host_templates.append({
                'objectType': 'SoftLayer_Hardware',
                'id': hardware_id
            })

    if virtual_guest_ids is not None:
        for virtual_guest_id in virtual_guest_ids:
            host_templates.append({
                'objectType': 'SoftLayer_Virtual_Guest',
                'id': virtual_guest_id
            })

    if ip_address_ids is not None:
        for ip_address_id in ip_address_ids:
            host_templates.append({
                'objectType': 'SoftLayer_Network_Subnet_IpAddress',
                'id': ip_address_id
            })

    if subnet_ids is not None:
        for subnet_id in subnet_ids:
            host_templates.append({
                'objectType': 'SoftLayer_Network_Subnet',
                'id': subnet_id
            })
    return host_templates


def get_package(manager, category_code):
    """Returns a product package based on type of storage.

    :param manager: The storage manager which calls this function.
    :param category_code: Category code of product package.
    :return: Returns a packaged based on type of storage.
    """

    _filter = utils.NestedDict({})
    _filter['categories']['categoryCode'] = (
        utils.query_filter(category_code))
    _filter['statusCode'] = utils.query_filter('ACTIVE')

    packages = manager.client.call(
        'Product_Package', 'getAllObjects',
        filter=_filter.to_dict(),
        mask='id,name,items[prices[categories],attributes]'
    )
    if len(packages) == 0:
        raise ValueError('No packages were found for %s' % category_code)
    if len(packages) > 1:
        raise ValueError('More than one package was found for %s'
                         % category_code)

    return packages[0]


def get_location_id(manager, location):
    """Returns location id

    :param manager: The storage manager which calls this function.
    :param location: Datacenter short name
    :return: Returns location id
    """
    loc_svc = manager.client['Location_Datacenter']
    datacenters = loc_svc.getDatacenters(mask='mask[longName,id,name]')
    for datacenter in datacenters:
        if datacenter['name'] == location:
            location = datacenter['id']
            return location
    raise ValueError('Invalid datacenter name specified.')


def find_price_by_category(package, price_category):
    """Find the price in the given package that has the specified category

    :param package: The AsAService, Enterprise, or Performance product package
    :param price_category: The price category code to search for
    :return: Returns the price for the given category, or an error if not found
    """
    for item in package['items']:
        price_id = _find_price_id(item['prices'], price_category)
        if price_id:
            return price_id

    raise ValueError("Could not find price with the category, %s" % price_category)


def find_ent_space_price(package, category, size, tier_level):
    """Find the space price for the given category, size, and tier

    :param package: The Enterprise (Endurance) product package
    :param category: The category of space (endurance, replication, snapshot)
    :param size: The size for which a price is desired
    :param tier_level: The endurance tier for which a price is desired
    :return: Returns the matching price, or an error if not found
    """
    if category == 'snapshot':
        category_code = 'storage_snapshot_space'
    elif category == 'replication':
        category_code = 'performance_storage_replication'
    else:  # category == 'endurance'
        category_code = 'performance_storage_space'

    level = ENDURANCE_TIERS.get(tier_level)

    for item in package['items']:
        if int(item['capacity']) != size:
            continue
        price_id = _find_price_id(item['prices'], category_code, 'STORAGE_TIER_LEVEL', level)
        if price_id:
            return price_id

    raise ValueError("Could not find price for %s storage space" % category)


def find_ent_endurance_tier_price(package, tier_level):
    """Find the price in the given package with the specified tier level

    :param package: The Enterprise (Endurance) product package
    :param tier_level: The endurance tier for which a price is desired
    :return: Returns the price for the given tier, or an error if not found
    """
    for item in package['items']:
        for attribute in item.get('attributes', []):
            if int(attribute['value']) == ENDURANCE_TIERS.get(tier_level):
                break
        else:
            continue

        price_id = _find_price_id(item['prices'], 'storage_tier_level')
        if price_id:
            return price_id

    raise ValueError("Could not find price for endurance tier level")


def find_endurance_tier_iops_per_gb(volume):
    """Find the tier for the given endurance volume (IOPS per GB)

    :param volume: The volume for which the tier level is desired
    :return: Returns a float value indicating the IOPS per GB for the volume
    """
    tier = volume['storageTierLevel']
    iops_per_gb = 0.25

    if tier == "LOW_INTENSITY_TIER":
        iops_per_gb = 0.25
    elif tier == "READHEAVY_TIER":
        iops_per_gb = 2
    elif tier == "WRITEHEAVY_TIER":
        iops_per_gb = 4
    elif tier == "10_IOPS_PER_GB":
        iops_per_gb = 10
    else:
        raise ValueError("Could not find tier IOPS per GB for this volume")

    return iops_per_gb


def find_perf_space_price(package, size):
    """Find the price in the given package with the specified size

    :param package: The Performance product package
    :param size: The storage space size for which a price is desired
    :return: Returns the price for the given size, or an error if not found
    """
    for item in package['items']:
        if int(item['capacity']) != size:
            continue

        price_id = _find_price_id(item['prices'], 'performance_storage_space')
        if price_id:
            return price_id

    raise ValueError("Could not find performance space price for this volume")


def find_perf_iops_price(package, size, iops):
    """Find the price in the given package with the specified size and iops

    :param package: The Performance product package
    :param size: The size of storage space for which an IOPS price is desired
    :param iops: The number of IOPS for which a price is desired
    :return: Returns the price for the size and IOPS, or an error if not found
    """
    for item in package['items']:
        if int(item['capacity']) != int(iops):
            continue

        price_id = _find_price_id(item['prices'], 'performance_storage_iops', 'STORAGE_SPACE', size)
        if price_id:
            return price_id

    raise ValueError("Could not find price for iops for the given volume")


def find_saas_endurance_space_price(package, size, tier_level):
    """Find the SaaS endurance storage space price for the size and tier

    :param package: The Storage As A Service product package
    :param size: The volume size for which a price is desired
    :param tier_level: The endurance tier for which a price is desired
    :return: Returns the price for the size and tier, or an error if not found
    """
    if tier_level != 0.25:
        tier_level = int(tier_level)
    key_name = f'STORAGE_SPACE_FOR_{tier_level}_IOPS_PER_GB'
    key_name = key_name.replace(".", "_")
    for item in package['items']:
        if key_name not in item['keyName']:
            continue

        if 'capacityMinimum' not in item or 'capacityMaximum' not in item:
            continue

        capacity_minimum = int(item['capacityMinimum'])
        capacity_maximum = int(item['capacityMaximum'])
        if size < capacity_minimum or size > capacity_maximum:
            continue

        price_id = _find_price_id(item['prices'], 'performance_storage_space')
        if price_id:
            return price_id

    raise ValueError("Could not find price for endurance storage space")


def find_saas_endurance_tier_price(package, tier_level):
    """Find the SaaS storage tier level price for the specified tier level

    :param package: The Storage As A Service product package
    :param tier_level: The endurance tier for which a price is desired
    :return: Returns the price for the given tier, or an error if not found
    """
    target_capacity = ENDURANCE_TIERS.get(tier_level)
    for item in package['items']:
        if 'itemCategory' not in item\
                or 'categoryCode' not in item['itemCategory']\
                or item['itemCategory']['categoryCode']\
                != 'storage_tier_level':
            continue

        if int(item['capacity']) != target_capacity:
            continue

        price_id = _find_price_id(item['prices'], 'storage_tier_level')
        if price_id:
            return price_id

    raise ValueError("Could not find price for endurance tier level")


def find_saas_perform_space_price(package, size):
    """Find the SaaS performance storage space price for the given size

    :param package: The Storage As A Service product package
    :param size: The volume size for which a price is desired
    :return: Returns the price for the size and tier, or an error if not found
    """
    for item in package['items']:
        if 'itemCategory' not in item\
                or 'categoryCode' not in item['itemCategory']\
                or item['itemCategory']['categoryCode']\
                != 'performance_storage_space':
            continue

        if 'capacityMinimum' not in item or 'capacityMaximum' not in item:
            continue

        capacity_minimum = int(item['capacityMinimum'])
        capacity_maximum = int(item['capacityMaximum'])
        if size < capacity_minimum or size > capacity_maximum:
            continue

        key_name = f'{capacity_minimum}_{capacity_maximum}_GBS'
        if item['keyName'] != key_name:
            continue
        price_id = _find_price_id(item['prices'], 'performance_storage_space')
        if price_id:
            return price_id

    raise ValueError("Could not find price for performance storage space")


def find_saas_perform_iops_price(package, size, iops):
    """Find the SaaS IOPS price for the specified size and iops

    :param package: The Storage As A Service product package
    :param size: The volume size for which a price is desired
    :param iops: The number of IOPS for which a price is desired
    :return: Returns the price for the size and IOPS, or an error if not found
    """
    for item in package['items']:
        if 'itemCategory' not in item\
                or 'categoryCode' not in item['itemCategory']\
                or item['itemCategory']['categoryCode']\
                != 'performance_storage_iops':
            continue

        if 'capacityMinimum' not in item or 'capacityMaximum' not in item:
            continue

        capacity_minimum = int(item['capacityMinimum'])
        capacity_maximum = int(item['capacityMaximum'])
        if iops < capacity_minimum or iops > capacity_maximum:
            continue

        price_id = _find_price_id(item['prices'], 'performance_storage_iops', 'STORAGE_SPACE', size)
        if price_id:
            return price_id

    raise ValueError("Could not find price for iops for the given volume")


def find_saas_snapshot_space_price(package, size, tier=None, iops=None):
    """Find the price in the SaaS package for the desired snapshot space size

    :param package: The product package of the endurance storage type
    :param size: The snapshot space size for which a price is desired
    :param tier: The tier of the volume for which space is being ordered
    :param iops: The IOPS of the volume for which space is being ordered
    :return: Returns the price for the given size, or an error if not found
    """
    if tier is not None:
        target_value = ENDURANCE_TIERS.get(tier)
        target_restriction_type = 'STORAGE_TIER_LEVEL'
    else:
        target_value = iops
        target_restriction_type = 'IOPS'

    for item in package['items']:
        if int(item['capacity']) != size:
            continue

        price_id = _find_price_id(item['prices'], 'storage_snapshot_space', target_restriction_type, target_value)
        if price_id:
            return price_id

    raise ValueError("Could not find price for snapshot space")


def find_saas_replication_price(package, tier=None, iops=None):
    """Find the price in the given package for the desired replicant volume

    :param package: The product package of the endurance storage type
    :param tier: The tier of the primary storage volume
    :param iops: The IOPS of the primary storage volume
    :return: Returns the replication price, or an error if not found
    """
    if tier is not None:
        target_value = ENDURANCE_TIERS.get(tier)
        target_item_keyname = 'REPLICATION_FOR_TIERBASED_PERFORMANCE'
        target_restriction_type = 'STORAGE_TIER_LEVEL'
    else:
        target_value = iops
        target_item_keyname = 'REPLICATION_FOR_IOPSBASED_PERFORMANCE'
        target_restriction_type = 'IOPS'

    for item in package['items']:
        if item['keyName'] != target_item_keyname:
            continue

        price_id = _find_price_id(
            item['prices'],
            'performance_storage_replication',
            target_restriction_type,
            target_value
        )
        if price_id:
            return price_id

    raise ValueError("Could not find price for replicant volume")


def find_snapshot_schedule_id(volume, snapshot_schedule_keyname):
    """Find the snapshot schedule ID for the given volume and keyname

    :param volume: The volume for which the snapshot ID is desired
    :param snapshot_schedule_keyname: The keyname of the snapshot schedule
    :return: Returns an int value indicating the volume's snapshot schedule ID
    """
    for schedule in volume['schedules']:
        if 'type' in schedule and 'keyname' in schedule['type']:
            if schedule['type']['keyname'] == snapshot_schedule_keyname:
                return schedule['id']

    raise ValueError("The given snapshot schedule ID was not found for the given storage volume")


def prepare_snapshot_order_object(manager, volume, capacity, tier, upgrade, iops):
    """Prepare the snapshot space order object for the placeOrder() method

    :param manager: The File or Block manager calling this function
    :param integer volume: The volume for which snapshot space is ordered
    :param integer capacity: The snapshot space size to order, in GB
    :param float tier: The tier level of the volume, in IOPS per GB (optional)
    :param boolean upgrade: Flag to indicate if this order is an upgrade
    :return: Returns the order object for the
             Product_Order service's placeOrder() method
    """
    # Ensure the storage volume has not been cancelled
    if 'billingItem' not in volume:
        raise exceptions.SoftLayerError('This volume has been cancelled; unable to order snapshot space')

    # Determine and validate the storage volume's billing item category
    billing_item_category_code = volume['billingItem']['categoryCode']
    if billing_item_category_code == 'storage_as_a_service':
        order_type_is_saas = True
    elif billing_item_category_code == 'storage_service_enterprise':
        order_type_is_saas = False
    else:
        raise exceptions.SoftLayerError(
            "Snapshot space cannot be ordered for a primary volume with a "
            "billing item category code of '%s'" % billing_item_category_code)

    # Use the volume's billing item category code to get the product package
    package = get_package(manager, billing_item_category_code)

    # Find prices based on the volume's type and billing item category
    if order_type_is_saas:  # 'storage_as_a_service' package
        volume_storage_type = volume['storageType']['keyName']
        if 'ENDURANCE' in volume_storage_type:
            if tier is None:
                tier = find_endurance_tier_iops_per_gb(volume)
            prices = [find_saas_snapshot_space_price(
                package, capacity, tier=tier)]
        elif 'PERFORMANCE' in volume_storage_type:
            if not _staas_version_is_v2_or_above(volume):
                raise exceptions.SoftLayerError(
                    "Snapshot space cannot be ordered for this performance "
                    "volume since it does not support Encryption at Rest.")

            prices = [find_saas_snapshot_space_price(
                package, capacity, iops=iops)]
        else:
            raise exceptions.SoftLayerError(
                "Storage volume does not have a valid storage type "
                "(with an appropriate keyName to indicate the "
                "volume is a PERFORMANCE or an ENDURANCE volume)")
    else:  # 'storage_service_enterprise' package
        if tier is None:
            tier = find_endurance_tier_iops_per_gb(volume)
        prices = [find_ent_space_price(package, 'snapshot', capacity, tier)]

    # Currently, these types are valid for snapshot space orders, whether
    # the base volume's order container was Enterprise or AsAService
    if upgrade:
        complex_type = 'SoftLayer_Container_Product_Order_'\
                       'Network_Storage_Enterprise_SnapshotSpace_Upgrade'
    else:
        complex_type = 'SoftLayer_Container_Product_Order_'\
                       'Network_Storage_Enterprise_SnapshotSpace'

    # Determine if hourly billing should be used
    hourly_billing_flag = utils.lookup(volume, 'billingItem', 'hourlyFlag')
    if hourly_billing_flag is None:
        hourly_billing_flag = False

    # Build and return the order object
    snapshot_space_order = {
        'complexType': complex_type,
        'packageId': package['id'],
        'prices': prices,
        'quantity': 1,
        'location': volume['billingItem']['location']['id'],
        'volumeId': volume['id'],
        'useHourlyPricing': hourly_billing_flag
    }

    return snapshot_space_order


def prepare_volume_order_object(manager, storage_type, location, size,
                                iops, tier, snapshot_size, service_offering,
                                volume_type, hourly_billing_flag=False):
    """Prepare the order object which is submitted to the placeOrder() method

    :param manager: The File or Block manager calling this function
    :param storage_type: "performance" or "endurance"
    :param location: Requested datacenter location name for the ordered volume
    :param size: Desired size of the volume, in GB
    :param iops: Number of IOPs for a "Performance" volume order
    :param tier: Tier level to use for an "Endurance" volume order
    :param snapshot_size: The size of snapshot space for the volume (optional)
    :param service_offering: Requested offering package to use for the order
    :param volume_type: The type of the volume to order ('file' or 'block')
    :param hourly_billing_flag: Billing type, monthly (False) or hourly (True)
    :return: Returns the order object for the
             Product_Order service's placeOrder() method
    """
    # Ensure the volume storage type is valid
    if storage_type != 'performance' and storage_type != 'endurance':
        raise exceptions.SoftLayerError(
            "Volume storage type must be either performance or endurance")

    # Find the ID for the requested location
    try:
        location_id = get_location_id(manager, location)
    except ValueError as ex:
        message = "Invalid datacenter name specified. Please provide the lower case short name (e.g.: dal09)"
        raise exceptions.SoftLayerError(message) from ex

    # Determine the category code to use for the order (and product package)
    order_type_is_saas, order_category_code = _get_order_type_and_category(
        service_offering,
        storage_type,
        volume_type
    )

    # Get the product package for the given category code
    package = get_package(manager, order_category_code)

    # Based on the storage type and product package, build up the complex type
    # and array of price codes to include in the order object
    base_type_name = 'SoftLayer_Container_Product_Order_Network_'
    if order_type_is_saas:
        complex_type = base_type_name + 'Storage_AsAService'
        if storage_type == 'performance':
            prices = [
                find_price_by_category(package, order_category_code),
                find_price_by_category(package, 'storage_' + volume_type),
                find_saas_perform_space_price(package, size),
                find_saas_perform_iops_price(package, size, iops)
            ]
            if snapshot_size is not None:
                prices.append(find_saas_snapshot_space_price(
                    package, snapshot_size, iops=iops))
        else:  # storage_type == 'endurance'
            prices = [
                find_price_by_category(package, order_category_code),
                find_price_by_category(package, 'storage_' + volume_type),
                find_saas_endurance_space_price(package, size, tier),
                find_saas_endurance_tier_price(package, tier)
            ]
            if snapshot_size is not None:
                prices.append(find_saas_snapshot_space_price(
                    package, snapshot_size, tier=tier))
    else:  # offering package is enterprise or performance
        if storage_type == 'performance':
            if volume_type == 'block':
                complex_type = base_type_name + 'PerformanceStorage_Iscsi'
            else:
                complex_type = base_type_name + 'PerformanceStorage_Nfs'
            prices = [
                find_price_by_category(package, order_category_code),
                find_perf_space_price(package, size),
                find_perf_iops_price(package, size, iops),
            ]
        else:  # storage_type == 'endurance'
            complex_type = base_type_name + 'Storage_Enterprise'
            prices = [
                find_price_by_category(package, order_category_code),
                find_price_by_category(package, 'storage_' + volume_type),
                find_ent_space_price(package, 'endurance', size, tier),
                find_ent_endurance_tier_price(package, tier),
            ]
            if snapshot_size is not None:
                prices.append(find_ent_space_price(
                    package, 'snapshot', snapshot_size, tier))

    # Build and return the order object
    order = {
        'complexType': complex_type,
        'packageId': package['id'],
        'prices': prices,
        'quantity': 1,
        'location': location_id,
        'useHourlyPricing': hourly_billing_flag
    }

    if order_type_is_saas:
        order['volumeSize'] = size
        if storage_type == 'performance':
            order['iops'] = iops

    return order


def _get_order_type_and_category(service_offering, storage_type, volume_type):
    if service_offering == 'storage_as_a_service':
        order_type_is_saas = True
        order_category_code = 'storage_as_a_service'
    elif service_offering == 'enterprise':
        order_type_is_saas = False
        if storage_type == 'endurance':
            order_category_code = 'storage_service_enterprise'
        else:
            raise exceptions.SoftLayerError(
                "The requested offering package, '%s', is not available for "
                "the '%s' storage type." % (service_offering, storage_type))
    elif service_offering == 'performance':
        order_type_is_saas = False
        if storage_type == 'performance':
            if volume_type == 'block':
                order_category_code = 'performance_storage_iscsi'
            else:
                order_category_code = 'performance_storage_nfs'
        else:
            raise exceptions.SoftLayerError(
                "The requested offering package, '%s', is not available for "
                "the '%s' storage type." % (service_offering, storage_type))
    else:
        raise exceptions.SoftLayerError(
            "The requested service offering package is not valid. "
            "Please check the available options and try again.")

    return order_type_is_saas, order_category_code


def prepare_replicant_order_object(manager, snapshot_schedule, location,
                                   tier, volume, volume_type, iops):
    """Prepare the order object which is submitted to the placeOrder() method

    :param manager: The File or Block manager calling this function
    :param snapshot_schedule: The primary volume's snapshot
                              schedule to use for replication
    :param location: The location for the ordered replicant volume
    :param tier: The tier (IOPS per GB) of the primary volume
    :param volume: The primary volume as a SoftLayer_Network_Storage object
    :param volume_type: The type of the primary volume ('file' or 'block')
    :return: Returns the order object for the
             Product_Order service's placeOrder() method
    """
    # Ensure the primary volume and snapshot space are not set for cancellation
    if 'billingItem' not in volume\
            or volume['billingItem'].get('cancellationDate'):
        raise exceptions.SoftLayerError(
            'This volume is set for cancellation; '
            'unable to order replicant volume')

    for child in volume['billingItem']['activeChildren']:
        if child['categoryCode'] == 'storage_snapshot_space'\
                and child.get('cancellationDate'):
            raise exceptions.SoftLayerError(
                'The snapshot space for this volume is set for '
                'cancellation; unable to order replicant volume')

    # Find the ID for the requested location
    try:
        location_id = get_location_id(manager, location)
    except ValueError as ex:
        message = "Invalid datacenter name specified. Please provide the lower case short name (e.g.: dal09)"
        raise exceptions.SoftLayerError(message) from ex

    # Get sizes and properties needed for the order
    volume_size = int(volume['capacityGb'])

    billing_item_category_code = volume['billingItem']['categoryCode']
    if billing_item_category_code == 'storage_as_a_service':
        order_type_is_saas = True
    elif billing_item_category_code == 'storage_service_enterprise':
        order_type_is_saas = False
    else:
        raise exceptions.SoftLayerError(
            "A replicant volume cannot be ordered for a primary volume with a "
            "billing item category code of '%s'" % billing_item_category_code)

    if 'snapshotCapacityGb' in volume:
        snapshot_size = int(volume['snapshotCapacityGb'])
    else:
        raise exceptions.SoftLayerError(
            "Snapshot capacity not found for the given primary volume")

    snapshot_schedule_id = find_snapshot_schedule_id(
        volume,
        'SNAPSHOT_' + snapshot_schedule
    )

    # Use the volume's billing item category code to get the product package
    package = get_package(manager, billing_item_category_code)

    # Find prices based on the primary volume's type and billing item category
    if order_type_is_saas:  # 'storage_as_a_service' package
        complex_type = 'SoftLayer_Container_Product_Order_'\
                       'Network_Storage_AsAService'
        volume_storage_type = volume['storageType']['keyName']
        if 'ENDURANCE' in volume_storage_type:
            volume_is_performance = False
            if tier is None:
                tier = find_endurance_tier_iops_per_gb(volume)
            prices = [
                find_price_by_category(package, billing_item_category_code),
                find_price_by_category(package, 'storage_' + volume_type),
                find_saas_endurance_space_price(package, volume_size, tier),
                find_saas_endurance_tier_price(package, tier),
                find_saas_snapshot_space_price(
                    package, snapshot_size, tier=tier),
                find_saas_replication_price(package, tier=tier)
            ]
        elif 'PERFORMANCE' in volume_storage_type:
            if not _staas_version_is_v2_or_above(volume):
                raise exceptions.SoftLayerError(
                    "A replica volume cannot be ordered for this performance "
                    "volume since it does not support Encryption at Rest.")
            volume_is_performance = True

            prices = [
                find_price_by_category(package, billing_item_category_code),
                find_price_by_category(package, 'storage_' + volume_type),
                find_saas_perform_space_price(package, volume_size),
                find_saas_perform_iops_price(package, volume_size, iops),
                find_saas_snapshot_space_price(
                    package, snapshot_size, iops=iops),
                find_saas_replication_price(package, iops=iops)
            ]
        else:
            raise exceptions.SoftLayerError(
                "Storage volume does not have a valid storage type "
                "(with an appropriate keyName to indicate the "
                "volume is a PERFORMANCE or an ENDURANCE volume)")
    else:  # 'storage_service_enterprise' package
        complex_type = 'SoftLayer_Container_Product_Order_'\
                       'Network_Storage_Enterprise'
        volume_is_performance = False
        if tier is None:
            tier = find_endurance_tier_iops_per_gb(volume)
        prices = [
            find_price_by_category(package, billing_item_category_code),
            find_price_by_category(package, 'storage_' + volume_type),
            find_ent_space_price(package, 'endurance', volume_size, tier),
            find_ent_endurance_tier_price(package, tier),
            find_ent_space_price(package, 'snapshot', snapshot_size, tier),
            find_ent_space_price(package, 'replication', volume_size, tier)
        ]

    # Determine if hourly billing should be used
    hourly_billing_flag = utils.lookup(volume, 'billingItem', 'hourlyFlag')
    if hourly_billing_flag is None:
        hourly_billing_flag = False

    # Build and return the order object
    replicant_order = {
        'complexType': complex_type,
        'packageId': package['id'],
        'prices': prices,
        'quantity': 1,
        'location': location_id,
        'originVolumeId': volume['id'],
        'originVolumeScheduleId': snapshot_schedule_id,
        'useHourlyPricing': hourly_billing_flag
    }

    if order_type_is_saas:
        replicant_order['volumeSize'] = volume_size
        if volume_is_performance:
            replicant_order['iops'] = iops

    return replicant_order


def prepare_duplicate_order_object(manager, origin_volume, iops, tier,
                                   duplicate_size, duplicate_snapshot_size,
                                   volume_type, hourly_billing_flag=False,
                                   dependent_duplicate=False):
    """Prepare the duplicate order to submit to SoftLayer_Product::placeOrder()

    :param manager: The File or Block manager calling this function
    :param origin_volume: The origin volume which is being duplicated
    :param iops: The IOPS for the duplicate volume (performance)
    :param tier: The tier level for the duplicate volume (endurance)
    :param duplicate_size: The requested size for the duplicate volume
    :param duplicate_snapshot_size: The size for the duplicate snapshot space
    :param volume_type: The type of the origin volume ('file' or 'block')
    :param hourly_billing_flag: Billing type, monthly (False) or hourly (True)
    :param dependent_duplicate: Duplicate type, normal (False) or dependent
        duplicate (True)
    :return: Returns the order object to be passed to the
             placeOrder() method of the Product_Order service
    """

    # Verify that the origin volume has not been cancelled
    if 'billingItem' not in origin_volume:
        raise exceptions.SoftLayerError(
            "The origin volume has been cancelled; "
            "unable to order duplicate volume")

    # Verify that the origin volume has snapshot space (needed for duplication)
    if isinstance(utils.lookup(origin_volume, 'snapshotCapacityGb'), str):
        origin_snapshot_size = int(origin_volume['snapshotCapacityGb'])
    else:
        raise exceptions.SoftLayerError(
            "Snapshot space not found for the origin volume. "
            "Origin snapshot space is needed for duplication.")

    # Obtain the datacenter location ID for the duplicate
    if isinstance(utils.lookup(origin_volume, 'billingItem',
                               'location', 'id'), int):
        location_id = origin_volume['billingItem']['location']['id']
    else:
        raise exceptions.SoftLayerError(
            "Cannot find origin volume's location")

    # Ensure the origin volume is STaaS v2 or higher
    # and supports Encryption at Rest
    if not _staas_version_is_v2_or_above(origin_volume):
        raise exceptions.SoftLayerError(
            "This volume cannot be duplicated since it "
            "does not support Encryption at Rest.")

    # If no specific snapshot space was requested for the duplicate,
    # use the origin snapshot space size
    if duplicate_snapshot_size is None:
        duplicate_snapshot_size = origin_snapshot_size

    # Use the origin volume size if no size was specified for the duplicate
    if duplicate_size is None:
        duplicate_size = origin_volume['capacityGb']

    # Get the appropriate package for the order
    # ('storage_as_a_service' is currently used for duplicate volumes)
    package = get_package(manager, 'storage_as_a_service')

    # Determine the IOPS or tier level for the duplicate volume, along with
    # the type and prices for the order
    origin_storage_type = origin_volume['storageType']['keyName']
    if 'PERFORMANCE' in origin_storage_type:
        volume_is_performance = True
        if iops is None:
            iops = int(origin_volume.get('provisionedIops', 0))
            if iops <= 0:
                raise exceptions.SoftLayerError("Cannot find origin volume's provisioned IOPS")
        # Set up the price array for the order
        prices = [
            find_price_by_category(package, 'storage_as_a_service'),
            find_price_by_category(package, 'storage_' + volume_type),
            find_saas_perform_space_price(package, duplicate_size),
            find_saas_perform_iops_price(package, duplicate_size, iops),
        ]
        # Add the price code for snapshot space as well, unless 0 GB was given
        if duplicate_snapshot_size > 0:
            prices.append(find_saas_snapshot_space_price(
                package, duplicate_snapshot_size, iops=iops))

    elif 'ENDURANCE' in origin_storage_type:
        volume_is_performance = False
        if tier is None:
            tier = find_endurance_tier_iops_per_gb(origin_volume)
        # Set up the price array for the order
        prices = [
            find_price_by_category(package, 'storage_as_a_service'),
            find_price_by_category(package, 'storage_' + volume_type),
            find_saas_endurance_space_price(package, duplicate_size, tier),
            find_saas_endurance_tier_price(package, tier),
        ]
        # Add the price code for snapshot space as well, unless 0 GB was given
        if duplicate_snapshot_size > 0:
            prices.append(find_saas_snapshot_space_price(
                package, duplicate_snapshot_size, tier=tier))

    else:
        raise exceptions.SoftLayerError(
            "Origin volume does not have a valid storage type "
            "(with an appropriate keyName to indicate the "
            "volume is a PERFORMANCE or an ENDURANCE volume)")

    duplicate_order = {
        'complexType': 'SoftLayer_Container_Product_Order_'
                       'Network_Storage_AsAService',
        'packageId': package['id'],
        'prices': prices,
        'volumeSize': duplicate_size,
        'quantity': 1,
        'location': location_id,
        'duplicateOriginVolumeId': origin_volume['id'],
        'useHourlyPricing': hourly_billing_flag
    }

    if volume_is_performance:
        duplicate_order['iops'] = iops

    if dependent_duplicate:
        duplicate_order['isDependentDuplicateFlag'] = 1

    return duplicate_order


def prepare_modify_order_object(manager, volume, new_iops, new_tier, new_size):
    """Prepare the modification order to submit to SoftLayer_Product::placeOrder()

    :param manager: The File or Block manager calling this function
    :param volume: The volume which is being modified
    :param new_iops: The new IOPS for the volume (performance)
    :param new_tier: The new tier level for the volume (endurance)
    :param new_size: The requested new size for the volume
    :return: Returns the order object to be passed to the placeOrder() method of the Product_Order service
    """

    # Verify that the origin volume has not been cancelled
    if 'billingItem' not in volume:
        raise exceptions.SoftLayerError("The volume has been cancelled; unable to modify volume.")

    # Ensure the origin volume is STaaS v2 or higher and supports Encryption at Rest
    if not _staas_version_is_v2_or_above(volume):
        raise exceptions.SoftLayerError("This volume cannot be modified since it does not support Encryption at Rest.")

    # Get the appropriate package for the order ('storage_as_a_service' is currently used for modifying volumes)
    package = get_package(manager, 'storage_as_a_service')

    # Based on volume storage type, ensure at least one volume property is being modified,
    # use current values if some are not specified, and lookup price codes for the order
    volume_storage_type = volume['storageType']['keyName']
    if 'PERFORMANCE' in volume_storage_type:
        volume_is_performance = True
        if new_size is None and new_iops is None:
            raise exceptions.SoftLayerError("A size or IOPS value must be given to modify this performance volume.")

        if new_size is None:
            new_size = volume['capacityGb']
        elif new_iops is None:
            new_iops = int(volume.get('provisionedIops', 0))
            if new_iops <= 0:
                raise exceptions.SoftLayerError("Cannot find volume's provisioned IOPS.")

        # Set up the prices array for the order
        prices = [
            find_price_by_category(package, 'storage_as_a_service'),
            find_saas_perform_space_price(package, new_size),
            find_saas_perform_iops_price(package, new_size, new_iops),
        ]

    elif 'ENDURANCE' in volume_storage_type:
        volume_is_performance = False
        if new_size is None and new_tier is None:
            raise exceptions.SoftLayerError("A size or tier value must be given to modify this endurance volume.")

        if new_size is None:
            new_size = volume['capacityGb']
        elif new_tier is None:
            new_tier = find_endurance_tier_iops_per_gb(volume)

        # Set up the prices array for the order
        prices = [
            find_price_by_category(package, 'storage_as_a_service'),
            find_saas_endurance_space_price(package, new_size, new_tier),
            find_saas_endurance_tier_price(package, new_tier),
        ]

    else:
        raise exceptions.SoftLayerError("Volume does not have a valid storage type (with an appropriate "
                                        "keyName to indicate the volume is a PERFORMANCE or an ENDURANCE volume).")

    modify_order = {
        'complexType': 'SoftLayer_Container_Product_Order_Network_Storage_AsAService_Upgrade',
        'packageId': package['id'],
        'prices': prices,
        'volume': {'id': volume['id']},
        'volumeSize': new_size
    }

    if volume_is_performance:
        modify_order['iops'] = new_iops

    return modify_order


def block_or_file(storage_type_keyname):
    """returns either 'block' or 'file'

    :param storage_type_keyname: the Network_Storage['storageType']['keyName']
    :returns: 'block' or 'file'
    """
    return 'block' if 'BLOCK_STORAGE' in storage_type_keyname else 'file'


def _has_category(categories, category_code):
    return any(
        True
        for category
        in categories
        if category['categoryCode'] == category_code
    )


def _staas_version_is_v2_or_above(volume):
    return int(volume['staasVersion']) > 1 and volume['hasEncryptionAtRest']


def _find_price_id(prices, category, restriction_type=None, restriction_value=None):
    for price in prices:
        # Only collect prices from valid location groups.
        if price['locationGroupId']:
            continue

        if restriction_type is not None and restriction_value is not None:
            if restriction_type != price['capacityRestrictionType']\
                    or restriction_value < int(price['capacityRestrictionMinimum'])\
                    or restriction_value > int(price['capacityRestrictionMaximum']):
                continue

        if not _has_category(price['categories'], category):
            continue

        return {'id': price['id']}



"""Download SSL certificate and key file."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Download SSL certificate and key file."""

    manager = SoftLayer.SSLManager(env.client)
    certificate = manager.get_certificate(identifier)

    write_cert(certificate['commonName'] + '.crt', certificate['certificate'])
    write_cert(certificate['commonName'] + '.key', certificate['privateKey'])

    if 'intermediateCertificate' in certificate:
        write_cert(certificate['commonName'] + '.icc',
                   certificate['intermediateCertificate'])

    if 'certificateSigningRequest' in certificate:
        write_cert(certificate['commonName'] + '.csr',
                   certificate['certificateSigningRequest'])


def write_cert(filename, content):
    """Writes certificate body to the given file path."""
    with open(filename, 'w', encoding="utf-8") as cert_file:
        cert_file.write(content)



"""Get storage details for a virtual server."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get storage details for a virtual server."""

    vsi = SoftLayer.VSManager(env.client)
    vsi_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')
    iscsi_storage_data = vsi.get_storage_details(vsi_id, "ISCSI")
    nas_storage_data = vsi.get_storage_details(vsi_id, "NAS")
    storage_credentials = vsi.get_storage_credentials(vsi_id)
    portable_storage = vsi.get_portable_storage(vsi_id)
    local_disks = vsi.get_local_disks(vsi_id)

    table_credentials = formatting.Table(['Username', 'Password', 'IQN'], title="Block Storage Details \n iSCSI")
    if storage_credentials:
        table_credentials.add_row([storage_credentials['credential']['username'],
                                   storage_credentials['credential']['password'],
                                   storage_credentials['name']])

    table_iscsi = formatting.Table(['LUN name', 'capacity', 'Target address', 'Location', 'Notes'])
    for iscsi in iscsi_storage_data:
        table_iscsi.add_row([iscsi['username'], iscsi['capacityGb'],
                             iscsi['serviceResourceBackendIpAddress'],
                             iscsi['allowedVirtualGuests'][0]['datacenter']['longName'],
                             iscsi.get('notes', None)])

    table_portable = formatting.Table(['Description', 'Capacity'], title="Portable Storage")
    for portable in portable_storage:
        table_portable.add_row([portable.get('description', None), portable.get('capacity', None)])

    table_nas = formatting.Table(['Volume name', 'capacity', 'Host Name', 'Location', 'Notes'],
                                 title="File Storage Details")
    for nas in nas_storage_data:
        table_nas.add_row([nas['username'], nas['capacityGb'],
                           nas['serviceResourceBackendIpAddress'],
                           nas['allowedVirtualGuests'][0]['datacenter']['longName'],
                           nas.get('notes', None)])

    table_local_disks = get_local_storage_table(local_disks)
    table_local_disks.title = "Other storage details"

    env.fout(table_credentials)
    env.fout(table_iscsi)
    env.fout(table_portable)
    env.fout(table_nas)
    env.fout(table_local_disks)


def get_local_type(disks):
    """Returns the virtual server local disk type.

    :param disks: virtual server local disks.
    """
    disk_type = 'System'
    if 'SWAP' in disks.get('diskImage', {}).get('description', []):
        disk_type = 'Swap'

    return disk_type


def get_local_storage_table(local_disks):
    """Returns a formatting local disk table

      :param local_disks: virtual server local disks.
      """
    table_local_disks = formatting.Table(['Type', 'Name', 'Drive', 'Capacity'])
    for disk in local_disks:
        if 'diskImage' in disk:
            table_local_disks.add_row([
                get_local_type(disk),
                disk['mountType'],
                disk['device'],
                f"{disk['diskImage']['capacity']} {disk['diskImage']['units']}"
            ])
    return table_local_disks



"""
    SoftLayer.object_storage
    ~~~~~~~~~~~~~~~~~~~~~~~~
    Object Storage Manager/helpers

    :license: MIT, see LICENSE for more details.
"""

from SoftLayer.exceptions import SoftLayerError
from SoftLayer import utils

LIST_ACCOUNTS_MASK = '''mask[
    id,username,notes,vendorName,serviceResource
]'''


class ObjectStorageManager(utils.IdentifierMixin, object):
    """Manager for SoftLayer Object Storage accounts.

    See product information here: https://www.ibm.com/cloud/object-storage

    :param SoftLayer.API.BaseClient client: the client instance

    """

    def __init__(self, client):
        self.client = client
        self.resolvers = [self._get_id_from_username]

    def list_accounts(self, object_mask=None, object_filter=None, limit=10):
        """Lists your object storage accounts."""
        object_mask = object_mask if object_mask else LIST_ACCOUNTS_MASK
        return self.client.call('Account',
                                'getHubNetworkStorage',
                                mask=object_mask,
                                filter=object_filter,
                                limit=limit)

    def list_endpoints(self, identifier):
        """Lists the known object storage endpoints."""

        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'getEndpoints', id=identifier)

    def create_credential(self, identifier):
        """Create object storage credential.

        :param int identifier: The object storage account identifier.

        """

        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'credentialCreate',
                                id=identifier)

    def delete_credential(self, identifier, credential_id=None):
        """Delete the object storage credential.

        :param int id: The object storage account identifier.
        :param int credential_id: The credential id to be deleted.

        """
        credential = {
            'id': credential_id
        }

        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'credentialDelete',
                                credential, id=identifier)

    def limit_credential(self, identifier):
        """Limit object storage credentials.

        :param int identifier: The object storage account identifier.

        """

        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'getCredentialLimit',
                                id=identifier)

    def list_credential(self, identifier):
        """List the object storage credentials.

        :param int identifier: The object storage account identifier.

        """

        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'getCredentials',
                                id=identifier)

    def _get_id_from_username(self, username):
        """Looks up a username's id

        :param string username: Username to lookup
        :returns: The id that matches username.
        """
        _mask = "mask[id,username]"
        _filter = {'hubNetworkStorage': {'username': utils.query_filter(username)}}
        account = self.list_accounts(_mask, _filter)
        if len(account) == 1:
            return [account[0]['id']]
        elif len(account) > 1:
            raise SoftLayerError(f"Multiple object storage accounts found with the name: {username}")
        else:
            raise SoftLayerError(f"Unable to find object storage account id for: {username}")



"""
    SoftLayer.CLI.columns
    ~~~~~~~~~~~~~~~~~~~~~~
    Utilties for user-defined columns

    :license: MIT, see LICENSE for more details.
"""
import click

from SoftLayer import utils

# pylint: disable=unused-argument


class Column(object):
    """Column desctribes an attribute and how to fetch/display it."""

    def __init__(self, name, path, mask=None):
        self.name = name
        self.path = path
        self.mask = mask

        # If the mask is not set explicitly, infer it from the path
        if self.mask is None and isinstance(path, (tuple, list)):
            self.mask = '.'.join(path)


class ColumnFormatter(object):
    """Maps each column using a function"""

    def __init__(self):
        self.columns = []
        self.column_funcs = []
        self.mask_parts = set()

    def add_column(self, column):
        """Add a new column along with a formatting function."""
        self.columns.append(column.name)
        self.column_funcs.append(column.path)

        if column.mask is not None:
            self.mask_parts.add(column.mask)

    def row(self, data):
        """Return a formatted row for the given data."""
        for column in self.column_funcs:
            if callable(column):
                yield column(data)
            else:
                yield utils.lookup(data, *column)

    def mask(self):
        """Returns a SoftLayer mask to fetch data needed for each column."""
        return ','.join(self.mask_parts)


def get_formatter(columns):
    """This function returns a callback to use with click options.

    The returned function parses a comma-separated value and returns a new
    ColumnFormatter.

    :param columns: a list of Column instances
    """

    column_map = dict((column.name, column) for column in columns)

    def validate(ctx, param, value):
        """Click validation function."""
        if value == '':
            raise click.BadParameter('At least one column is required.')

        formatter = ColumnFormatter()
        for column in [col.strip() for col in value.split(',')]:
            if column in column_map:
                formatter.add_column(column_map[column])
            else:
                formatter.add_column(Column(column, column.split('.')))

        return formatter

    return validate



"""Delete a security group."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('securitygroup_id')
@environment.pass_env
def cli(env, securitygroup_id):
    """Deletes the given security group"""
    mgr = SoftLayer.NetworkManager(env.client)
    if not mgr.delete_securitygroup(securitygroup_id):
        raise exceptions.CLIAbort("Failed to delete security group")



"""List Placement Groups"""

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.vs_placement import PlacementManager as PlacementManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """List placement groups."""
    manager = PlacementManager(env.client)
    result = manager.list()
    table = formatting.Table(
        ["Id", "Name", "Backend Router", "Rule", "Guests", "Created"],
        title="Placement Groups"
    )
    for group in result:
        table.add_row([
            utils.lookup(group, 'id'),
            utils.lookup(group, 'name'),
            utils.lookup(group, 'backendRouter', 'hostname'),
            utils.lookup(group, 'rule', 'name'),
            utils.lookup(group, 'guestCount'),
            utils.lookup(group, 'createDate')
        ])

    env.fout(table)



"""Delete bandwidth pool."""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer import BandwidthManager
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Delete bandwidth pool."""

    manager = BandwidthManager(env.client)
    manager.delete_pool(identifier)
    env.fout(f"Bandwidth pool {identifier} has been scheduled for deletion.")



"""Edit hardware details."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--domain', '-D', help="Domain portion of the FQDN")
@click.option('--userfile', '-F', type=click.Path(exists=True, readable=True, resolve_path=True),
              help="Read userdata from file")
@click.option('--tag', '-g', multiple=True,
              help="Tags to set or empty string to remove all")
@click.option('--hostname', '-H', help="Host portion of the FQDN")
@click.option('--userdata', '-u', help="User defined metadata string")
@click.option('--public-speed', default=None, type=click.Choice(['0', '10', '100', '1000', '10000', '-1']),
              help="Public port speed. -1 is best speed available")
@click.option('--private-speed', default=None, type=click.Choice(['0', '10', '100', '1000', '10000', '-1']),
              help="Private port speed. -1 is best speed available")
@click.option('--redundant', is_flag=True, default=False, help="The desired state of redundancy for the interface(s)")
@click.option('--degraded', is_flag=True, default=False, help="The desired state of degraded for the interface(s)")
@environment.pass_env
def cli(env, identifier, domain, userfile, tag, hostname, userdata, public_speed, private_speed, redundant, degraded):
    """Edit hardware details."""

    if userdata and userfile:
        raise exceptions.ArgumentError(
            '[-u | --userdata] not allowed with [-F | --userfile]')

    data = {
        'hostname': hostname,
        'domain': domain,
    }

    if userdata:
        data['userdata'] = userdata
    elif userfile:
        with open(userfile, 'r', encoding="utf-8") as userfile_obj:
            data['userdata'] = userfile_obj.read()

    if tag:
        data['tags'] = ','.join(tag)

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')

    if not mgr.edit(hw_id, **data):
        raise exceptions.CLIAbort("Failed to update hardware")

    if public_speed is not None:
        if redundant:
            mgr.change_port_speed(hw_id, True, int(public_speed), 'redundant')
        if degraded:
            mgr.change_port_speed(hw_id, True, int(public_speed), 'degraded')
        if not redundant and not degraded:
            raise exceptions.CLIAbort("Failed to update hardwar")

    if private_speed is not None:
        if redundant:
            mgr.change_port_speed(hw_id, False, int(private_speed), 'redundant')
        if degraded:
            mgr.change_port_speed(hw_id, False, int(private_speed), 'degraded')
        if not redundant and not degraded:
            raise exceptions.CLIAbort("Failed to update hardware")



"""Import zone based off a BIND zone file."""
# :license: MIT, see LICENSE for more details.
import re

import click

import SoftLayer
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers

RECORD_REGEX = re.compile(r"""^((?P<domain>(([\w-]+|\*)(\.)?)*|\@)?\s+
                               (?P<ttl>\d+)?\s+
                               (?P<class>\w+)?)?\s+
                               (?P<type>\w+)\s+
                               (?P<data>.*)""", re.X)
RECORD_FMT = "type={type}, record={record}, data={data}, ttl={ttl}"


@click.command(cls=SLCommand)
@click.argument('zonefile', type=click.Path(exists=True, readable=True, resolve_path=True))
@click.option('--dry-run', is_flag=True, help="Don't actually create records")
@environment.pass_env
def cli(env, zonefile, dry_run):
    """Import zone based off a BIND zone file.

    Example::
        slcli dns import ~/ibm.com.txt
        This command imports zone and its resource records from file: ~/ibm.com.txt
"""

    manager = SoftLayer.DNSManager(env.client)
    with open(zonefile, encoding="utf-8") as zone_f:
        zone_contents = zone_f.read()

    zone, records, bad_lines = parse_zone_details(zone_contents)

    click.secho("Parsed: zone=%s" % zone)
    for record in records:
        click.secho("Parsed: %s" % RECORD_FMT.format(**record), fg="green")

    for line in bad_lines:
        click.secho("Unparsed: %s" % line, fg="yellow")

    if dry_run:
        return

    # Find zone id or create the zone if it doesn't exist
    try:
        zone_id = helpers.resolve_id(manager.resolve_ids, zone,
                                     name='zone')
    except exceptions.CLIAbort:
        zone_id = manager.create_zone(zone)['id']
        click.secho(click.style("Created: %s" % zone, fg='green'))

    # Attempt to create each record
    for record in records:
        try:
            manager.create_record(zone_id, record['record'], record['type'], record['data'], record['ttl'])

            click.secho(click.style("Created: %s" % RECORD_FMT.format(**record), fg='green'))
        except SoftLayer.SoftLayerAPIError as ex:
            click.secho(click.style("Failed: %s" % RECORD_FMT.format(**record), fg='red'))
            click.secho(click.style(str(ex), fg='red'))

    click.secho(click.style("Finished", fg='green'))


def parse_zone_details(zone_contents):
    """Parses a zone file into python data-structures."""
    records = []
    bad_lines = []
    zone_lines = [line.strip() for line in zone_contents.split('\n')]

    zone_search = re.search(r'^\$ORIGIN (?P<zone>.*)\.', zone_lines[0])
    if zone_search:
        zone = zone_search.group('zone')
    else:
        raise exceptions.ArgumentError("Invalid Zone File")

    for line in zone_lines[1:]:
        record_search = re.search(RECORD_REGEX, line)
        if record_search is None:
            bad_lines.append(line)
            continue

        name = record_search.group('domain')
        # The API requires we send a host, although bind allows a blank
        # entry. @ is the same thing as blank
        if name is None:
            name = "@"

        ttl = record_search.group('ttl')
        # we don't do anything with the class
        # domain_class = domainSearch.group('class')
        record_type = record_search.group('type').upper()
        data = record_search.group('data')

        # the dns class doesn't support weighted MX records yet, so we chomp
        # that part out.
        if record_type == "MX":
            record_search = re.search(r'(?P<weight>\d+)\s+(?P<data>.*)', data)
            data = record_search.group('data')

        # This will skip the SOA record bit. And any domain that gets
        # parsed oddly.
        if record_type == 'IN':
            bad_lines.append(line)
            continue

        records.append({
            'record': name,
            'type': record_type,
            'data': data,
            'ttl': ttl,
        })

    return zone, records, bad_lines



"""List Users."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting

TWO_FACTO_AUTH = 'externalBindingCount'
CLASSIC_API_KEYS = 'apiAuthenticationKeyCount'

COLUMNS = [
    column_helper.Column('id', ('id',)),
    column_helper.Column('username', ('username',)),
    column_helper.Column('email', ('email',)),
    column_helper.Column('displayName', ('displayName',)),
    column_helper.Column('status', ('userStatus', 'name')),
    column_helper.Column('hardwareCount', ('hardwareCount',)),
    column_helper.Column('virtualGuestCount', ('virtualGuestCount',)),
    column_helper.Column('2FA', (TWO_FACTO_AUTH,)),
    column_helper.Column('classicAPIKey', (CLASSIC_API_KEYS,)),
    column_helper.Column('vpn', ('sslVpnAllowedFlag',))
]

DEFAULT_COLUMNS = [
    'id',
    'username',
    'email',
    'displayName',
    '2FA',
    'classicAPIKey',
    'vpn'
]


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--columns',
              callback=column_helper.get_formatter(COLUMNS),
              help=f"Columns to display. [options: {', '.join(column.name for column in COLUMNS)}]",
              default=','.join(DEFAULT_COLUMNS),
              show_default=True)
@environment.pass_env
def cli(env, columns):
    """List Users."""

    mgr = SoftLayer.UserManager(env.client)
    users = mgr.list_users()

    table = formatting.Table(columns.columns)
    for user in users:
        user = _yes_format(user, [TWO_FACTO_AUTH, CLASSIC_API_KEYS, 'sslVpnAllowedFlag'])
        table.add_row([value or formatting.blank()
                       for value in columns.row(user)])

    env.fout(table)


def _yes_format(user, keys):
    """Changes all dictionary values to yes whose keys are in the list. """
    for key in keys:
        if user.get(key):
            user[key] = 'yes'
    return user



"""Search for SoftLayer Resources by simple phrase."""
# :license: MIT, see LICENSE for more details.


import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.search import SearchManager

object_types = {'ticket': 'SoftLayer_Ticket',
                'firewall': 'SoftLayer_Network_Vlan_Firewall',
                'vlan': 'SoftLayer_Network_Vlan',
                'subnet': 'SoftLayer_Network_Subnet_IpAddress',
                'delivery': 'SoftLayer_Network_Application_Delivery_Controller',
                'dedicated': 'SoftLayer_Virtual_DedicatedHost',
                'log': 'SoftLayer_Event_Log',
                'hardware': 'SoftLayer_Hardware',
                'virtual': 'SoftLayer_Virtual_Guest'}


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('query', nargs=-1)
@click.option('--types', is_flag=True, default=False, is_eager=True, help="Display searchable types.")
@click.option('--advanced', is_flag=True, help="Calls the AdvancedSearh API.")
@click.help_option('--help', '-h')
@environment.pass_env
def cli(env, query, types, advanced):
    """Perform a query against the SoftLayer search database.

    Read More: https://sldn.softlayer.com/reference/services/SoftLayer_Search/search/
    Examples:

        slcli search test.com
        slcli search _objectType:SoftLayer_Virtual_Guest test.com
        slcli -vvv search _objectType:SoftLayer_Hardware hostname:testibm --advanced
    """

    # Checks to make sure we have at least 1 query.
    def check_opt(list_opt=None):
        check = False
        for input_ in list_opt:
            if input_:
                check = True
                break

        return check

    list_opt = [query, types, advanced]

    if not check_opt(list_opt):
        raise click.UsageError('Search query must be provided')

    search = SearchManager(env.client)
    # query is in array, so we need to convert it to a normal string for the API
    query = " ".join(query)
    if types:
        search_types = search.get_object_types()

        table = formatting.Table(["Name", "Properties"])
        table.align = "r"

        for type_object in search_types:
            prop_table = formatting.Table(["Property", "Sortable"])
            prop_table.align = "l"
            for prop in type_object.get('properties'):
                prop_table.add_row([prop.get('name'), prop.get('sortableFlag')])
            table.add_row([type_object.get('name'), prop_table])

        env.fout(table)
        return
    if advanced:
        result = search.advanced(query)

        env.fout(formatting.iter_to_table(result))
    else:
        result = search.search(query)
        env.fout(formatting.iter_to_table(result))
